# coding=utf-8
"""Module for calculating sun positions and visualizing the sun path."""
from __future__ import division

from .location import Location
from .dt import DateTime
from .analysisperiod import AnalysisPeriod
from .compass import Compass

from ladybug_geometry.geometry3d.pointvector import Vector3D, Point3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.arc import Arc3D
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polyline import Polyline2D

import datetime as py_datetime
import math
import sys
if (sys.version_info > (3, 0)):  # python 3
    xrange = range


class Sunpath(object):
    """Calculate sun positions and visualize the sun path

    Args:
        latitude: A number between -90 and 90 for the latitude of the location
            in degrees. (Default: 0 for the equator)
        longitude: A number between -180 and 180 for the longitude of the location
            in degrees (Default: 0 for the prime meridian)
        time_zone: A number representing the time zone of the location for the
            sun path. Typically, this value is an integer, assuming that a
            standard time zone is used but this value can also be a decimal
            for the purposes of modeling location-specific solar time.
            The time zone should follow the epw convention and should be
            between -12 and +14, where 0 is at Greenwich, UK, positive values
            are to the East of Greenwich and negative values are to the West.
            If None, this value will be set to solar time using the Sunpath's
            longitude. (Default: None).
        north_angle: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East (Default: 0).
        daylight_saving_period: An analysis period for daylight saving time.
            If None, no daylight saving time will be used. (Default: None)

    Properties:
        * latitude
        * longitude
        * time_zone
        * north_angle
        * daylight_saving_period
        * is_leap_year

    Usage:

    .. code-block:: python

        import ladybug.sunpath as sunpath
        # initiate sunpath
        sp = sunpath.Sunpath(50)
        sun = sp.calculate_sun(1, 1, 12) # calculate sun data for Jan 1 at noon
        print(sun.azimuth, sun.altitude)
    """

    __slots__ = ('_longitude', '_latitude', '_north_angle', '_time_zone',
                 '_daylight_saving_period', '_is_leap_year')
    PI = math.pi

    def __init__(self, latitude=0, longitude=0, time_zone=None, north_angle=0,
                 daylight_saving_period=None):
        """Init sunpath.
        """
        self.latitude = latitude
        self.longitude = longitude
        self.time_zone = time_zone
        self.north_angle = north_angle
        self.daylight_saving_period = daylight_saving_period
        self._is_leap_year = False

    @classmethod
    def from_location(cls, location, north_angle=0, daylight_saving_period=None):
        """Create a sun path from a ladybug.location.Location."""
        location = Location.from_location(location)
        return cls(location.latitude, location.longitude,
                   location.time_zone, north_angle, daylight_saving_period)

    @property
    def latitude(self):
        """Get or set a number between -90 and 90 for the latitude in degrees."""
        return math.degrees(self._latitude)

    @latitude.setter
    def latitude(self, value):
        self._latitude = math.radians(float(value))
        assert -self.PI / 2 <= self._latitude <= self.PI / 2, \
            'latitude value should be between -90 and 90. Got {}.'.format(value)
        if self._latitude == self.PI / 2:  # prevent math domain errors
            self._latitude = self._latitude - 1e-9
        if self._latitude == -self.PI / 2:  # prevent math domain errors
            self._latitude = self._latitude + 1e-9

    @property
    def longitude(self):
        """Get or set a number between -180 and 180 for the longitude in degrees.

        Note that you will also likely want to update the time zone of the
        Sunpath if this value is set to somethign far from its original value.
        """
        return math.degrees(self._longitude)

    @longitude.setter
    def longitude(self, value):
        self._longitude = math.radians(float(value))
        assert -self.PI <= self._longitude <= self.PI, \
            'longitude value should be between -180 and 180. Got {}.'.format(value)

    @property
    def time_zone(self):
        """Get or set the time zone as a number between -12 and + 14.

        Setting this property to None will automatically set the Sunpath to use
        solar time for the Sunpath's longitude.
        """
        return self._time_zone

    @time_zone.setter
    def time_zone(self, tz):
        self._time_zone = self.longitude / 15 if tz is None else float(tz)
        assert -12 <= self._time_zone <= 14, \
            'Time zone must be between -12 and +14. Got {}.'.format(self._time_zone)

    @property
    def north_angle(self):
        """Get or set a number between -360 and 360 for the north_angle in degrees."""
        return math.degrees(self._north_angle)

    @north_angle.setter
    def north_angle(self, value):
        self._north_angle = math.radians(float(value))
        assert -self.PI * 2 <= self._north_angle <= self.PI * 2, \
            'north_angle value should be between -360 and 360. Got {}.'.format(value)

    @property
    def is_leap_year(self):
        """Get or set a boolean to indicate is sunpath calculated for a leap year."""
        return self._is_leap_year

    @is_leap_year.setter
    def is_leap_year(self, value):
        """set sunpath to be calculated for a leap year."""
        self._is_leap_year = bool(value)

    @property
    def daylight_saving_period(self):
        """Get or set an AnalysisPeriod for the daylight saving period.

        Note that the st_hour and end_hour of the AnalysisPeriod will be interpreted
        as indicating the time of day when the shift in clocks happen. So setting the
        st_hour and end_hour to 2 will ensure that the time shift happens at 2 AM on
        the dates specified. Note that this is different than how the hour inputs of
        AnalysisPeriods typically work, where the st_hour and end_hour apply to every
        day of the analysis period.

        If None, no daylight saving period will be used.
        """
        return self._daylight_saving_period

    @daylight_saving_period.setter
    def daylight_saving_period(self, value):
        if value is not None:
            assert isinstance(value, AnalysisPeriod), \
                'Daylight saving period should be an AnalysisPeriod not %s' % type(value)
        self._daylight_saving_period = value

    def is_daylight_saving_hour(self, datetime):
        """Check if a datetime is within the daylight saving time."""
        if not self.daylight_saving_period:
            return False
        return self.daylight_saving_period.st_time.moy <= datetime.moy < \
            self.daylight_saving_period.end_time.moy

    def calculate_sun(self, month, day, hour, is_solar_time=False):
        """Get Sun data for an hour of the year.

        Args:
            month: An integer between 1 and 12.
            day: An integer between 1 and 31.
            hour: A positive number between 0 and 23. This can be a decimal value
                to yield a solar position in between hours (eg. 12.5).
            is_solar_time: A boolean to indicate if the input hour is in solar
                time. (Default: False)

        Returns:
            A sun object for the input month, day, hour.
        """
        datetime = DateTime(month, day, *self._calculate_hour_and_minute(hour),
                            leap_year=self.is_leap_year)
        return self.calculate_sun_from_date_time(datetime, is_solar_time)

    def calculate_sun_from_hoy(self, hoy, is_solar_time=False):
        """Get Sun data for an hour of the year.

        Args:
            hoy: A number for the hour of the year. This can be a decimal value
                to yield a solar position in between hours (eg. 12.5).
            is_solar_time: A boolean to indicate if the input hoy is in solar
                time. (Default: False)

        Returns:
            A sun object for the input hoy.
        """
        datetime = DateTime.from_hoy(hoy, self.is_leap_year)
        return self.calculate_sun_from_date_time(datetime, is_solar_time)

    def calculate_sun_from_moy(self, moy, is_solar_time=False):
        """Get Sun data for a minute of the year.

        Args:
            moy: An integer for the minute of the year.
            is_solar_time: A boolean to indicate if the input moy is in solar
                time. (Default: False)

        Returns:
            A sun object for the input moy.
        """
        datetime = DateTime.from_moy(moy, self.is_leap_year)
        return self.calculate_sun_from_date_time(datetime, is_solar_time)

    def calculate_sun_from_date_time(self, datetime, is_solar_time=False):
        """Get Sun for a specific datetime.

        This code is originally written by Trygve Wastvedt (Trygve.Wastvedt@gmail.com)
        based on (NOAA) and modified by Chris Mackey and Mostapha Roudsari.

        Args:
            datetime: Ladybug datetime.
            is_solar_time: A boolean to indicate if the input hour is in solar
                time. (Default: False)

        Returns:
            A sun object for the input datetime.
        """
        # TODO(mostapha): This should be more generic and based on a method
        if datetime.year != 2016 and self.is_leap_year:
            datetime = DateTime(datetime.month, datetime.day, datetime.hour,
                                datetime.minute, True)

        # compute solar geometry
        sol_dec, eq_of_time = self._calculate_solar_geometry(datetime)

        # get the correct mintue of the day for which solar position is to be computed
        try:
            hour = datetime.float_hour
        except AttributeError:  # native Python datetime; try to compute manually
            hour = datetime.hour + datetime.minute / 60.0
        is_daylight_saving = self.is_daylight_saving_hour(datetime)
        hour = hour - 1 if is_daylight_saving else hour  # spring forward!
        sol_time = self._calculate_solar_time(hour, eq_of_time, is_solar_time) * 60

        # degrees for the angle between solar noon and the current time.
        hour_angle = sol_time / 4 + 180 if sol_time < 0 else sol_time / 4 - 180

        # radians for the zenith and degrees for altitude
        zenith = math.acos(math.sin(self._latitude) * math.sin(sol_dec) +
                           math.cos(self._latitude) * math.cos(sol_dec) *
                           math.cos(math.radians(hour_angle)))
        altitude = 90 - math.degrees(zenith)

        # approx atmospheric refraction used to correct the altitude
        if altitude > 85:
            atmos_refraction = 0
        elif altitude > 5:
            atmos_refraction = 58.1 / math.tan(math.radians(altitude)) - \
                0.07 / (math.tan(math.radians(altitude))) ** 3 + \
                0.000086 / (math.tan(math.radians(altitude))) ** 5
        elif altitude > -0.575:
            atmos_refraction = 1735 + altitude * \
                (-518.2 + altitude * (103.4 + altitude * (-12.79 + altitude * 0.711)))
        else:
            atmos_refraction = -20.772 / math.tan(math.radians(altitude))
        
        atmos_refraction /= 3600
        altitude += atmos_refraction

        # azimuth in degrees
        az_init = ((math.sin(self._latitude) * math.cos(zenith)) - math.sin(sol_dec)) / \
                    (math.cos(self._latitude) * math.sin(zenith))
        try:
            if hour_angle > 0:
                azimuth = (math.degrees(math.acos(az_init)) + 180) % 360
            else:
                azimuth = (540 - math.degrees(math.acos(az_init))) % 360
        except ValueError:  # perfect solar noon yields math domain error
            azimuth = 180

        # create the sun for this hour
        return Sun(datetime, altitude, azimuth, is_solar_time, is_daylight_saving,
                   self.north_angle)

    def calculate_sunrise_sunset(self, month, day, depression=0.5334,
                                 is_solar_time=False):
        """Calculate sunrise, noon and sunset.

        Args:
            month: Integer for the month in which sunrise and sunset are computed.
            day: Integer for the day of the month in which sunrise and sunset
                are computed.
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                compute sunrise/sunset as the time when the edge of the sun begins
                to touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will compute sunrise/sunset as the time when the sun
                just finishes passing the horizon (actual sunset). Setting it
                to 0.833 will compute the apparent sunrise/sunset, accounting for
                atmospheric refraction. Setting this to 6 will compute sunrise/sunset
                as the beginning/end of civil twilight. Setting this to 12 will
                compute sunrise/sunset as the beginning/end of nautical twilight.
                Setting this to 18 will compute sunrise/sunset as the beginning/end
                of astronomical twilight. (Default: 0.5334).
            is_solar_time: A boolean to indicate if the output datetimes for sunrise,
                noon and sunset should be in solar time as opposed to the time zone
                of this Sunpath. (Default: False)

        Return:
            A dictionary. Keys are ("sunrise", "noon", "sunset"). Values are
            datetimes that correspond to these moments. Note that some values
            may be None if there is no sunrise or sunset on the specified day.
        """
        datetime = DateTime(month, day, hour=12, leap_year=self.is_leap_year)
        return self.calculate_sunrise_sunset_from_datetime(
            datetime, depression, is_solar_time)

    def calculate_sunrise_sunset_from_datetime(self, datetime, depression=0.5334,
                                               is_solar_time=False):
        """Calculate sunrise, sunset and noon for a day of year.
        
        Args:
            datetime: A ladybug DateTime object to indicate the month and day for
                which sunrise and sunset are computed.
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                compute sunrise/sunset as the time when the edge of the sun begins
                to touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will compute sunrise/sunset as the time when the sun
                just finishes passing the horizon (actual sunset). Setting it
                to 0.833 will compute the apparent sunrise/sunset, accounting for
                atmospheric refraction. Setting this to 6 will compute sunrise/sunset
                as the beginning/end of civil twilight. Setting this to 12 will
                compute sunrise/sunset as the beginning/end of nautical twilight.
                Setting this to 18 will compute sunrise/sunset as the beginning/end
                of astronomical twilight. (Default: 0.5334).
            is_solar_time: A boolean to indicate if the output datetimes for sunrise,
                noon and sunset should be in solar time as opposed to the time zone
                of this Sunpath. (Default: False)

        Return:
            A dictionary. Keys are ("sunrise", "noon", "sunset"). Values are
            datetimes that correspond to these moments. Note that some values
            may be None if there is no sunrise or sunset on the specified day.
        """
        # TODO(mostapha): This should be more generic and based on a method
        if datetime.year != 2016 and self.is_leap_year:
            datetime = DateTime(datetime.month, datetime.day, datetime.hour,
                                datetime.minute, True)
        sol_dec, eq_of_time = self._calculate_solar_geometry(datetime)
        # calculate sunrise and sunset hour
        if is_solar_time:
            noon = .5
        else:
            noon = (720 - 4 * self.longitude - eq_of_time + self.time_zone * 60) / 1440.

        try:
            sunrise_hour_angle = self._calculate_sunrise_hour_angle(
                sol_dec, math.radians(depression))
        except ValueError:
            # no sunrise/sunset on this day (eg. arctic circle in summer/winter)
            noon = 24 * noon
            return {
                "sunrise": None,
                "noon": DateTime(datetime.month, datetime.day,
                                 *self._calculate_hour_and_minute(noon),
                                 leap_year=self.is_leap_year),
                "sunset": None
            }
        else:
            sunrise = noon - sunrise_hour_angle * 4 / 1440.0
            sunset = noon + sunrise_hour_angle * 4 / 1440.0
            noon = 24 * noon
            sunrise = 24 * sunrise
            sunset = 24 * sunset

            return {
                "sunrise": DateTime(datetime.month, datetime.day,
                                    *self._calculate_hour_and_minute(sunrise),
                                    leap_year=self.is_leap_year),
                "noon": DateTime(datetime.month, datetime.day,
                                 *self._calculate_hour_and_minute(noon),
                                 leap_year=self.is_leap_year),
                "sunset": DateTime(datetime.month, datetime.day,
                                   *self._calculate_hour_and_minute(sunset),
                                   leap_year=self.is_leap_year)
            }

    def analemma_suns(self, time, daytime_only=False, is_solar_time=False):
        """Get an array of Suns that represent an analemma for a single time of day.

        Args:
            time: A ladybug Time object for the specific time of day to make
                the analemma.
            daytime_only: A boolean to note whether only daytime suns should
                be included in the resulting array. Note that this can
                result in a completely empty array. (Default: False)
            is_solar_time: A boolean to indicate if the output analemmas should
                be for solar hours instead of the hours of the sunpath time
                zone. (Default: False)

        Returns:
            An array of suns representing an analemma for a specific time.
            Analemmas will each have 12 suns for the 12 months of the year
            if daytime_only is False.
        """
        analemma = []
        for mon in range(1, 13):
            dat_tim = DateTime(mon, 21, time.hour, time.minute)
            analemma.append(self.calculate_sun_from_date_time(dat_tim, is_solar_time))
        if daytime_only:  # filter out the nighttime sun positions
            analemma = [sun for sun in analemma if sun.is_during_day]
        return analemma

    def hourly_analemma_suns(self, daytime_only=False, is_solar_time=False):
        """Get a nested array of Suns with one sub-array for each hourly analemma.

        Args:
            daytime_only: A boolean to note whether only daytime suns should
                be included in the resulting arrays. Note that this will likely
                result in completely empty arrays for some hours. (Default: False)
            is_solar_time: A boolean to indicate if the output analemmas should
                be for solar hours instead of the hours of the sunpath time
                zone. (Default: False)
        
        Returns:
            An array of 24 arrays with each sub-array representing an analemma.
            Analemmas will each have 12 suns for the 12 months of the year
        """
        analemmas = []  # list of polylines
        for hr in range(24):
            analem = []
            for mon in range(1, 13):
                dat_tim = DateTime(mon, 21, hr)
                analem.append(self.calculate_sun_from_date_time(dat_tim, is_solar_time))
            analemmas.append(analem)
        if daytime_only:  # filter out the nighttime sun positions
            for i, analem in enumerate(analemmas):
                analemmas[i] = [sun for sun in analem if sun.is_during_day]
        return analemmas

    def hourly_analemma_polyline3d(self, origin=Point3D(), radius=100,
                                   daytime_only=True, is_solar_time=False):
        """Get an array of ladybug_geometry Polyline3D for hourly analemmas.

        Args:
            origin: A ladybug_geometry Point3D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether only the daytime hours should
                be represented in the output Polyline3D. (Default: True)
            is_solar_time: A boolean to indicate if the output analemmas should
                be for solar hours instead of the hours of the sunpath time
                zone. (Default: False)
        
        Returns:
            An array of ladybug_geometry Polyline3D with at least one polyline
            for each analemma.
        """
        analemmas = []  # list of polylines
        analem_suns = self.hourly_analemma_suns(is_solar_time=is_solar_time)
        for analem in analem_suns:
            pts = []
            for sun in analem:
                pts.append(sun.position_3d(origin, radius))
            pts.append(pts[0])  # ensure that the Polyline3D is closed
            analemmas.append(Polyline3D(pts, interpolated=True))
        if not daytime_only:  # no need to further process the analemmas
            return analemmas

        # extract only the daytime portion of the analemmas
        daytime_analemmas = []
        plane_o = Plane(o=origin)
        for analem in analemmas:
            if analem.min.z > origin.z:  # fast check for analemmas above the plane
                daytime_analemmas.append(analem)
            elif analem.max.z < origin.z:  # fast check for analemmas below the plane
                pass
            else:  # split the Polylines with a plane at the input origin
                split_lines = analem.split_with_plane(plane_o)
                day_lines = []
                first_pt = None
                for pl in split_lines:
                    if isinstance(pl, Polyline3D) and pl.center.z > origin.z:
                        day_lines.append(pl)
                    elif isinstance(pl, LineSegment3D) and pl.midpoint.z > origin.z: 
                        try:  # last line segment in intersection; append it to the first
                            new_pl = Polyline3D((pl.p1,) + day_lines[0].vertices, True)
                            day_lines[0] = new_pl
                        except IndexError:  # first line segment in the intersection
                            first_pt = pl.p2
                if first_pt is not None and len(day_lines) > 0:
                    new_pl = Polyline3D(day_lines[0].vertices + (first_pt,), True)
                    day_lines[0] = new_pl
                daytime_analemmas.extend(day_lines)
        return daytime_analemmas

    def hourly_analemma_polyline2d(self, projection='Orthographic', origin=Point2D(),
                                   radius=100, daytime_only=True, is_solar_time=False):
        """Get an array of ladybug_geometry Polyline2D for hourly analemmas.

        Args:
            projection: Text for the name of the projection to use from the sky
                dome hemisphere to the 2D plane. (Default: 'Orthographic'). Choose 
                from the following:
                    * Orthographic
                    * Stereographic
            origin: A ladybug_geometry Point2D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether only the daytime hours should
                be represented in the output Polyline2D. (Default: True)
            is_solar_time: A boolean to indicate if the output analemmas should
                be for solar hours instead of the hours of the sunpath time
                zone. (Default: False)
        
        Returns:
            An array of ladybug_geometry Polyline2D with at least one polyline for
            each analemma.
        """
        # compute the analemmas in 3D space
        o_3d = Point3D(origin.x, origin.y, 0)
        plines_3d = self.hourly_analemma_polyline3d(o_3d, radius, daytime_only,
                                                    is_solar_time)
        return self._project_polyline_to_2d(plines_3d, projection, radius, o_3d)

    def day_arc3d(self, month, day, origin=Point3D(), radius=100, daytime_only=True,
                  depression=0.5334):
        """Get a ladybug_geometry Arc3D for the path taken by the sun on a single day.

        Args:
            month: Integer for the month in which sunrise and sunset are computed.
            day: Integer for the day of the month in which sunrise and sunset
                are computed.
            origin: A ladybug_geometry Point3D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether None should be returned if
                the sun never rises above the horizon on the the input day.
                For example, in the arctic circle in winter. (Default: True)
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                return an arc that ends when the edge of the sun begins to
                touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will return an arc that ends right at the horizon
                (actual sunset). Setting it to 0.833 will compute the apparent
                sunrise/sunset, accounting for atmospheric refraction. Setting to
                6 will return an arc that ends at the beginning/end of civil twilight.
                Setting to 12 will return an arc that ends at the beginning/end
                of nautical twilight. Setting to 18 will return an arc that
                ends at the beginning/end of astronomical twilight. (Default: 0.5334)

        Returns:
            An Arc3D for the path of the sun taken over the course of a day. Will be
            None if daytime_only is True and the sun is completely below the horizon
            for the entire day.
        """
        # get the sunrise, noon, and sunset time
        riseset_dict = self.calculate_sunrise_sunset(month, day, depression)

        # create the arcs from the sun positions
        if riseset_dict['sunrise'] is None:  # no sunrise; add a full circle
            noon = self.calculate_sun_from_date_time(riseset_dict['noon'])
            if daytime_only and noon.sun_vector.z > 0:  # night time
                return None
            sun6am = self.calculate_sun(month, day, 6)
            sun6pm = self.calculate_sun(month, day, 18)
            pts = [sun.position_3d(origin, radius) for sun in (sun6am, noon, sun6pm)]
            return Arc3D.from_start_mid_end(pts[0], pts[1], pts[2], circle=True)
        else:  # create an arc that respects the depression
            rise_noon_set_suns = (
                self.calculate_sun_from_date_time(riseset_dict['sunrise']),
                self.calculate_sun_from_date_time(riseset_dict['noon']),
                self.calculate_sun_from_date_time(riseset_dict['sunset']))
            positions = (sun.position_3d(origin, radius) for sun in rise_noon_set_suns)
            return Arc3D.from_start_mid_end(*positions)

    def day_polyline2d(self, month, day, projection='Orthographic', origin=Point2D(),
                       radius=100, daytime_only=True, depression=0.5334):
        """Get a Polyline2D for the path taken by the sun on a single day.

        Args:
            month: Integer for the month in which sunrise and sunset are computed.
            day: Integer for the day of the month in which sunrise and sunset
                are computed.
            projection: Text for the name of the projection to use from the sky
                dome hemisphere to the 2D plane. (Default: 'Orthographic'). Choose 
                from the following:
                    * Orthographic
                    * Stereographic
            origin: A ladybug_geometry Point2D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether None should be returned if
                the sun never rises above the horizon on the the input day.
                For example, in the arctic circle in winter. (Default: True)
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                return an arc that ends when the edge of the sun begins to
                touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will return an arc that ends right at the horizon
                (actual sunset). Setting it to 0.833 will compute the apparent
                sunrise/sunset, accounting for atmospheric refraction. Setting to
                6 will return an arc that ends at the beginning/end of civil twilight.
                Setting to 12 will return an arc that ends at the beginning/end
                of nautical twilight. Setting to 18 will return an arc that
                ends at the beginning/end of astronomical twilight. (Default: 0.5334)

        Returns:
            A Polyline2D for the path of the sun taken over the course of a day. Will be
            None if daytime_only is True and the sun is completely below the horizon
            for the entire day.
        """
        # compute the daily arc in 3D space
        o_3d = Point3D(origin.x, origin.y, 0)
        arc_3d = self.day_arc3d(month, day, o_3d, radius, daytime_only, depression)
        if arc_3d is not None:
            pline_3d = arc_3d.to_polyline(10, interpolated=True)  # convert to polyline
            return self._project_polyline_to_2d([pline_3d], projection, radius, o_3d)[0]

    def monthly_day_arc3d(self, origin=Point3D(), radius=100, daytime_only=True,
                          depression=0.5334):
        """Get an array of Arc3Ds for the path taken by the sun on the 21st of each month.

        Args:
            origin: A ladybug_geometry Point3D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether arcs should be excluded from the
                output array if the sun never rises above the horizon on the the day.
                For example, in the arctic circle in winter. (Default: True)
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                return an arc that ends when the edge of the sun begins to
                touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will return an arc that ends right at the horizon
                (actual sunset). Setting it to 0.833 will compute the apparent
                sunrise/sunset, accounting for atmospheric refraction. Setting to
                6 will return an arc that ends at the beginning/end of civil twilight.
                Setting to 12 will return an arc that ends at the beginning/end
                of nautical twilight. Setting to 18 will return an arc that
                ends at the beginning/end of astronomical twilight. (Default: 0.5334)

        Returns:
            An array of ladybug_geometry Arc3D with an arc for the 21st of each month.
        """
        day_arcs = []
        for mon in range(1, 13):
            arc = self.day_arc3d(mon, 21, origin, radius, daytime_only, depression)
            if arc is not None:
                day_arcs.append(arc)
        return day_arcs

    def monthly_day_polyline2d(self, projection='Orthographic', origin=Point2D(),
                               radius=100, daytime_only=True, depression=0.5334):
        """Get an array of Polyline2Ds for the sun path on the 21st of each month.

        Args:
            projection: Text for the name of the projection to use from the sky
                dome hemisphere to the 2D plane. (Default: 'Orthographic'). Choose 
                from the following:
                    * Orthographic
                    * Stereographic
            origin: A ladybug_geometry Point2D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
            daytime_only: A boolean to note whether arcs should be excluded from the
                output array if the sun never rises above the horizon on the the day.
                For example, in the arctic circle in winter. (Default: True)
            depression: An angle in degrees indicating the additional period
                before/after the edge of the sun has passed the horizon where
                the sun is still considered up. Setting this value to 0 will
                return an arc that ends when the edge of the sun begins to
                touch the horizon. Setting it to the angular diameter of the
                sun (0.5334) will return an arc that ends right at the horizon
                (actual sunset). Setting it to 0.833 will compute the apparent
                sunrise/sunset, accounting for atmospheric refraction. Setting to
                6 will return an arc that ends at the beginning/end of civil twilight.
                Setting to 12 will return an arc that ends at the beginning/end
                of nautical twilight. Setting to 18 will return an arc that
                ends at the beginning/end of astronomical twilight. (Default: 0.5334)

        Returns:
            An array of ladybug_geometry Polyline2D with a polyline for the 21st
            of each month.
        """
        # compute the daily arcs in 3D space
        o_3d = Point3D(origin.x, origin.y, 0)
        arcs_3d = self.monthly_day_arc3d(o_3d, radius, daytime_only, depression)
        plines_3d = [arc.to_polyline(10, interpolated=True) for arc in arcs_3d]
        return self._project_polyline_to_2d(plines_3d, projection, radius, o_3d)

    def _calculate_solar_geometry(self, datetime):
        """Calculate parameters related to solar geometry for an hour of the year.

        Attributes:
            datetime: A Ladybug datetime

        Returns:
            A tuple with two values

            - sol_dec: Solar declination in radians. Declination is analogous to
                latitude on Earth's surface, and measures an angular displacement
                north or south from the projection of Earth's equator on the
                celestial sphere to the location of a celestial body.

            - eq_of_time: Equation of time in minutes. This is an astronomical
                term accounting for changes in the time of solar noon for a given
                location over the course of a year. Earth's elliptical orbit and
                Kepler's law of equal areas in equal times are the culprits
                behind this phenomenon.
        """
        year, month, day, hour, minute = \
            datetime.year, datetime.month, datetime.day,  datetime.hour, datetime.minute

        julian_day = self._days_from_010119(year, month, day) + 2415018.5 + \
            round((minute + hour * 60) / 1440.0, 2) - (float(self.time_zone) / 24)

        julian_century = (julian_day - 2451545) / 36525

        # degrees
        geom_mean_long_sun = (280.46646 + julian_century *
                              (36000.76983 + julian_century * 0.0003032)
                              ) % 360
        # degrees
        geom_mean_anom_sun = 357.52911 + julian_century * \
            (35999.05029 - 0.0001537 * julian_century)

        eccent_orbit = 0.016708634 - julian_century * \
            (0.000042037 + 0.0000001267 * julian_century)

        sun_eq_of_ctr = math.sin(
            math.radians(geom_mean_anom_sun)) * \
            (1.914602 - julian_century * (0.004817 + 0.000014 * julian_century)
             ) +\
            math.sin(math.radians(2 * geom_mean_anom_sun)) * \
            (0.019993 - 0.000101 * julian_century) + \
            math.sin(math.radians(3 * geom_mean_anom_sun)) * \
            0.000289

        # degrees
        sun_true_long = geom_mean_long_sun + sun_eq_of_ctr

        # degrees
        sun_app_long = sun_true_long - 0.00569 - 0.00478 * \
            math.sin(math.radians(125.04 - 1934.136 * julian_century))

        # degrees
        mean_obliq_ecliptic = 23 + \
            (26 + ((21.448 - julian_century * (46.815 + julian_century *
                                               (0.00059 - julian_century *
                                                0.001813)))) / 60) / 60

        # degrees
        oblique_corr = mean_obliq_ecliptic + 0.00256 * \
            math.cos(math.radians(125.04 - 1934.136 * julian_century))

        # RADIANS
        sol_dec = math.asin(math.sin(math.radians(oblique_corr)) * \
            math.sin(math.radians(sun_app_long)))

        var_y = math.tan(math.radians(oblique_corr / 2)) * \
            math.tan(math.radians(oblique_corr / 2))

        # minutes
        eq_of_time = 4 \
            * math.degrees(
                var_y * math.sin(2 * math.radians(geom_mean_long_sun)) -
                2 * eccent_orbit * math.sin(math.radians(geom_mean_anom_sun)) +
                4 * eccent_orbit * var_y *
                math.sin(math.radians(geom_mean_anom_sun)) *
                math.cos(2 * math.radians(geom_mean_long_sun)) -
                0.5 * (var_y ** 2) *
                math.sin(4 * math.radians(geom_mean_long_sun)) -
                1.25 * (eccent_orbit ** 2) *
                math.sin(2 * math.radians(geom_mean_anom_sun))
            )

        return sol_dec, eq_of_time

    def _calculate_sunrise_hour_angle(self, solar_dec, depression):
        """Calculate hour angle for sunrise time in degrees.
        
        Args:
            solar_dec: Solar declination in radians.
            depression: Depression in radians.
        """
        hour_angle_arg = math.degrees(math.acos(
            math.cos(self.PI / 2 + depression) /
            (math.cos(self._latitude) * math.cos(solar_dec)) -
            math.tan(self._latitude) * math.tan(solar_dec)
        ))

        return hour_angle_arg

    def _calculate_solar_time(self, hour, eq_of_time, is_solar_time):
        """Calculate Solar time for an hour."""
        if is_solar_time:
            return hour

        return ((hour * 60 + eq_of_time + 4 * math.degrees(self._longitude) -
                 60 * self.time_zone) % 1440) / 60

    def _calculate_solar_time_by_doy(self, hour, doy):
        """This is how radiance calculates solar time.

        This is a place holder and should be validated against _calculate_solar_time.
        """
        raise NotImplementedError()
        return (0.170 * math.sin((4 * self.PI / 373) * (doy - 80)) -
                0.129 * math.sin((2 * self.PI / 355) * (doy - 8)) +
                12 * (-(15 * self.time_zone) - self.longitude) / self.PI)

    @staticmethod
    def _calculate_hour_and_minute(float_hour):
        """Calculate hour and minutes as integers from a float hour."""
        hour = int(float_hour)
        minute = int(round((float_hour - int(float_hour)) * 60))

        if minute == 60:
            return hour + 1, 0
        else:
            return hour, minute

    @staticmethod
    def _days_from_010119(year, month, day):
        """Calculate the number of days from 01-01-1900 to the provided date.
        
        Args:
            year: Integer. The year in the date
            month: Integer. The month in the date
            day: Integer. The day in the date

        Returns:
            The number of days since 01-01-1900 to the provided date
        """
        def is_leap_year(year):
            """Determine whether a year is a leap year over the past centuries."""
            return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

        if year == 2017:  # fast check for the most common year used in ladybug
            days_in_preceding_years = 42734
        elif year == 2016:  # fast check for the 2nd most common year in ladybug
            days_in_preceding_years = 42368
        else:  # compute the number of days using math to figure out leap years
            years = range(1900, year)  # list of years from 1900
            # Number of days in a year are 366 if it is a leap year
            days_in_year = []
            for item in years:
                if is_leap_year(item):
                    days_in_year.append(366)
                else:
                    days_in_year.append(365)
            # Making the total of all the days in preceding years
            days_in_preceding_years = 0
            for days in days_in_year:
                days_in_preceding_years += days

        # get the total of all the days in preceding months in the same year
        month_array = AnalysisPeriod.NUMOFDAYSEACHMONTHLEAP if is_leap_year(year) \
            else AnalysisPeriod.NUMOFDAYSEACHMONTH
        days_in_preceding_months = 0
        for i in range(month - 1):
            days_in_preceding_months += month_array[i]

        return days_in_preceding_years + days_in_preceding_months + day + 1

    @staticmethod
    def _project_polyline_to_2d(plines_3d, projection, radius, origin_3d):
        """Project an array of Polyline3D into 2D space.
        
        Args:
            plines_3d: An array of Polyline3D to be projected to 2D space.
            projection: Text for the name of the projection to use from the sky
                dome hemisphere to the 2D plane. Choose from the following:
                    * Orthographic
                    * Stereographic
            origin_3d: Point3D for the origin around which projection will occur.
        """
        plines_2d = []
        if projection.title() == 'Orthographic':
            for pl in plines_3d:
                pts = [Compass.point3d_to_orthographic(pt) for pt in pl.vertices]
                plines_2d.append(Polyline2D(pts, True))
        elif projection.title() == 'Stereographic':
            for pline in plines_3d:
                pts = [Compass.point3d_to_stereographic(pt, radius, origin_3d)
                        for pt in pline.vertices]
                plines_2d.append(Polyline2D(pts, True))
        else:
            raise ValueError('Projection "{}" is not supported.'.format(projection))
        return plines_2d

    def __repr__(self):
        """Sunpath representation."""
        return "Sunpath (lat:{}, lon:{}, time zone:{})".format(
            self.latitude, self.longitude, self.time_zone)


class Sun(object):
    """An object representing a single Sun.

    Args:
        datetime: A DateTime that represents the datetime for this sun_vector
        altitude: Solar Altitude in degrees.
        azimuth: Solar Azimuth in degrees.
        is_solar_time: Boolean indicating if the datetime represents a solar time.
        is_daylight_saving: A Boolean indicating if the datetime is calculated
            for a daylight saving period
        north_angle: North angle of the sunpath in degrees. This is only used to
            adjust the sun_vector and does not affect the sun altitude or azimuth.

    Properties:
        * datetime
        * north_angle
        * hoy
        * altitude
        * azimuth
        * altitude_in_radians
        * azimuth_in_radians
        * is_solar_time
        * is_daylight_saving
        * data
        * is_during_day
        * sun_vector
        * sun_vector_reversed
    """

    __slots__ = ('_datetime', '_altitude', '_azimuth', '_is_solar_time',
                 '_is_daylight_saving', '_north_angle', '_data',
                 '_sun_vector', '_sun_vector_reversed')
    PI = math.pi

    def __init__(self, datetime, altitude, azimuth, is_solar_time,
                 is_daylight_saving, north_angle, data=None):
        """Init sun."""
        assert isinstance(datetime, py_datetime.datetime), \
            'datetime must be a DateTime (not {})'.format(type(datetime))
        self._datetime = datetime  # read-only

        assert -90 <= altitude <= 90, \
            'altitude({}) must be between {} and {}.' \
            .format(altitude, -self.PI, self.PI)
        self._altitude = altitude  # read-only

        assert -360 <= azimuth <= 360, \
            'azimuth({}) should be between {} and {}.' \
            .format(azimuth, -self.PI, self.PI)
        self._azimuth = azimuth  # read-only
    
        self._is_solar_time = is_solar_time
        self._is_daylight_saving = is_daylight_saving
        self._north_angle = north_angle
        self.data = data  # place holder for metadata

        self._sun_vector, self._sun_vector_reversed = self._calculate_sun_vector()

    @property
    def datetime(self):
        """Return datetime."""
        return self._datetime

    @property
    def north_angle(self):
        """Return north angle for +YAxis."""
        return self._north_angle

    @property
    def hoy(self):
        """Return Hour of the year."""
        return self._datetime.hoy

    @property
    def altitude(self):
        """Return solar altitude in degrees."""
        return self._altitude

    @property
    def azimuth(self):
        """Return solar azimuth in degrees.

        This value is the same regardless of what the north_angle is.
        """
        return self._azimuth

    @property
    def azimuth_from_y_axis(self):
        """Return solar azimuth in degrees with respect to the Y-axis of the scene.

        This value will change as the north_angle is changed.
        """
        return  self._azimuth + self._north_angle

    @property
    def altitude_in_radians(self):
        """Return solar altitude in radians."""
        return math.radians(self._altitude)

    @property
    def azimuth_in_radians(self):
        """Return solar azimuth in radians."""
        return math.radians(self._azimuth)

    @property
    def is_solar_time(self):
        """Return a Boolean that indicates is datetime is solar time."""
        return self._is_solar_time

    @property
    def is_daylight_saving(self):
        """Return a Boolean that indicates is datetime is solar time."""
        return self._is_daylight_saving

    @property
    def data(self):
        """Get or set metadata to this sun position.
        
        No particular data type is enforced for this metadata but a dictionary
        is recommended so that it can be extended for multiple properties.
        """
        return self._data

    @data.setter
    def data(self, d):
        self._data = d

    @property
    def is_during_day(self):
        """Boolean to note if this sun position is during day."""
        # sun vector is flipped to look to the center
        return self.sun_vector.z <= 0

    @property
    def sun_vector(self):
        """A ladybug_geometry Vector3D representing the vector for this sun.

        Note that daytime sun vectors point downward (z will be negative).
        """
        return self._sun_vector

    @property
    def sun_vector_reversed(self):
        """A ladybug_geomtry Vector3D representing the reversed vector for this sun.

        Daytime sun_vector_reversed point upward (z will be positive).
        """
        return self._sun_vector_reversed
    
    def position_3d(self, origin=Point3D(), radius=100):
        """Get a Point3D for the position of this sun on a sunpath.

        Args:
            origin: A ladybug_geometry Point3D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.
        
        Returns:
            A Point3D for the position of this sun on a sunpath.
        """
        return Point3D(self.sun_vector_reversed.x * radius + origin.x,
                       self.sun_vector_reversed.y * radius + origin.y,
                       self.sun_vector_reversed.z * radius + origin.z)

    def position_2d(self, projection='Orthographic', origin=Point2D(), radius=100):
        """Get a Point2D for the position of this sun on a sunpath.

        Args:
            projection: Text for the name of the projection to use from the sky
                dome hemisphere to the 2D plane. (Default: 'Orthographic'). Choose 
                from the following:
                    * Orthographic
                    * Stereographic
            origin: A ladybug_geometry Point2D to note the center of the sun path.
            radius: A number to note the radius of the sunpath.

        Returns:
            A Point2D for the position of this sun on a sunpath.
        """
        o_3d = Point3D(origin.x, origin.y, 0)
        if projection.title() == 'Orthographic':
            return Compass.point3d_to_orthographic(self.position_3d(o_3d, radius))
        elif projection.title() == 'Stereographic':
            return Compass.point3d_to_stereographic(
                self.position_3d(o_3d, radius), radius, o_3d)
        else:
            raise ValueError('Projection "{}" is not supported.'.format(projection))

    def _calculate_sun_vector(self):
        """Calculate sun vector for this sun."""
        x_axis = Vector3D(1., 0., 0.)
        north_vector = Vector3D(0., 1., 0.)

        # rotate north vector based on azimuth, altitude, and north
        _sun_vector_reversed = north_vector \
            .rotate(x_axis, self.altitude_in_radians) \
            .rotate_xy(-self.azimuth_in_radians)
        if self.north_angle != 0:
            _sun_vector_reversed = _sun_vector_reversed.rotate_xy(
                math.radians(self.north_angle))

        # reverse the vector
        _sun_vector = _sun_vector_reversed.reverse()
        return _sun_vector, _sun_vector_reversed

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.datetime, self.altitude, self.azimuth, self.is_solar_time,
                self.is_daylight_saving, self.north_angle)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Sun) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Sun representation."""
        return "Sun at {} (x:{}, y:{}, z:{})".format(
            self.datetime,
            self.sun_vector.x,
            self.sun_vector.y,
            self.sun_vector.z
        )



# coding=utf-8
from ladybug.stat import STAT
from ladybug.ddy import DDY
from ladybug.designday import ASHRAEClearSky, ASHRAETau

from pytest import approx
import os


def test_import_stat():
    """Test import standard stat."""
    relative_path = './tests/fixtures/stat/chicago.stat'
    abs_path = os.path.abspath(relative_path)

    stat_rel = STAT(relative_path)
    stat = STAT(abs_path)

    # Test imports don't break
    assert stat.file_path == abs_path
    assert stat_rel.file_path == os.path.normpath(relative_path)
    assert isinstance(stat, STAT)


def test_stat_location():
    """Test the location within the stat object."""
    relative_path = './tests/fixtures/stat/tokyo.stat'
    stat = STAT(relative_path)

    # Test accuracy of import
    assert len(stat._header) == 10
    assert stat.location.city == 'TOKYO HYAKURI'
    assert stat.location.country == 'JPN'
    assert stat.location.source == 'IWEC Data'
    assert stat.location.station_id == '477150'
    assert stat.location.latitude == approx(36.18, rel=1e-1)
    assert stat.location.longitude == approx(140.42, rel=1e-1)
    assert stat.location.time_zone == 9
    assert stat.location.elevation == 35

    assert stat.monthly_tau_beam == \
        [0.341, 0.392, 0.48, 0.601, 0.654, 0.632, 0.553, 0.546, 0.473,
         0.423, 0.388, 0.338]
    assert stat.monthly_tau_diffuse == \
        [2.214, 1.999, 1.765, 1.546, 1.501, 1.591, 1.811, 1.827, 2.024,
         2.083, 2.105, 2.267]
    assert stat.ashrae_climate_zone == '4A'
    assert stat.koppen_climate_zone == 'Cfa'


def test_annual_heating_design_days():
    """Test the annual heating design days within the stat object."""
    relative_path = './tests/fixtures/stat/chicago.stat'
    stat = STAT(relative_path)

    ann_hdd_96 = stat.annual_heating_design_day_996
    ann_hdd_90 = stat.annual_heating_design_day_990

    assert ann_hdd_96.day_type == 'WinterDesignDay'
    assert ann_hdd_96.name == '99.6% Heating Design Day for Chicago Ohare Intl Ap'
    assert ann_hdd_96.analysis_period.st_month == \
        ann_hdd_96.analysis_period.end_month == \
        ann_hdd_96.sky_condition.date.month == 1
    assert ann_hdd_96.analysis_period.st_day == \
        ann_hdd_96.analysis_period.end_day == 21
    assert ann_hdd_96.humidity_condition.humidity_type == 'Wetbulb'
    assert ann_hdd_96.dry_bulb_condition.dry_bulb_max == \
        ann_hdd_96.humidity_condition.humidity_value == -20
    assert ann_hdd_96.dry_bulb_condition.dry_bulb_range == 0
    assert ann_hdd_96.wind_condition.wind_speed == 4.9
    assert ann_hdd_96.wind_condition.wind_direction == 270
    assert isinstance(ann_hdd_96.sky_condition, ASHRAEClearSky)
    assert ann_hdd_96.sky_condition.clearness == 0

    assert ann_hdd_90.name == '99.0% Heating Design Day for Chicago Ohare Intl Ap'
    assert ann_hdd_90.dry_bulb_condition.dry_bulb_max == \
        ann_hdd_90.humidity_condition.humidity_value == -16.6


def test_annual_cooling_design_days():
    """Test the annual cooling design days within the stat object."""
    relative_path = './tests/fixtures/stat/tokyo.stat'
    stat = STAT(relative_path)
    ann_cdd_04 = stat.annual_cooling_design_day_004
    ann_cdd_10 = stat.annual_cooling_design_day_010

    assert ann_cdd_04.day_type == 'SummerDesignDay'
    assert ann_cdd_04.name == '0.4% Cooling Design Day for TOKYO HYAKURI'
    assert ann_cdd_04.analysis_period.st_month == \
        ann_cdd_04.analysis_period.end_month == \
        ann_cdd_04.sky_condition.date.month == 8
    assert ann_cdd_04.analysis_period.st_day == \
        ann_cdd_04.analysis_period.end_day == 21
    assert ann_cdd_04.humidity_condition.humidity_type == 'Wetbulb'
    assert ann_cdd_04.dry_bulb_condition.dry_bulb_max == 32.1
    assert ann_cdd_04.humidity_condition.humidity_value == 26.0
    assert ann_cdd_04.dry_bulb_condition.dry_bulb_range == 7.7
    assert ann_cdd_04.wind_condition.wind_speed == 4.8
    assert ann_cdd_04.wind_condition.wind_direction == 210
    assert isinstance(ann_cdd_04.sky_condition, ASHRAETau)
    assert ann_cdd_04.sky_condition.tau_b == 0.546
    assert ann_cdd_04.sky_condition.tau_d == 1.827

    assert ann_cdd_10.name == '1.0% Cooling Design Day for TOKYO HYAKURI'
    assert ann_cdd_10.dry_bulb_condition.dry_bulb_max == 30.9
    assert ann_cdd_10.humidity_condition.humidity_value == 25.8


def test_monthly_cooling_design_days():
    """Test the monthly cooling design days within the stat object."""
    relative_path = './tests/fixtures/stat/chicago.stat'
    stat = STAT(relative_path)

    m_ddy_050 = stat.monthly_cooling_design_days_050
    m_ddy_100 = stat.monthly_cooling_design_days_100
    m_ddy_020 = stat.monthly_cooling_design_days_020
    m_ddy_004 = stat.monthly_cooling_design_days_004

    assert len(m_ddy_050) == len(m_ddy_100) == len(m_ddy_020) == \
        len(m_ddy_004) == 12

    ddy_path = './tests/fixtures/ddy/chicago_monthly.ddy'
    monthly_ddy = DDY(stat.location, m_ddy_050)
    monthly_ddy.save(ddy_path)


def test_typical_extreme_weeks():
    """Test the typical and extreme weeks within the stat object."""
    relative_path = './tests/fixtures/stat/chicago.stat'
    stat = STAT(relative_path)

    extreme_cold = stat.extreme_cold_week
    extreme_hot = stat.extreme_hot_week
    typical_winter = stat.typical_winter_week
    typical_spring = stat.typical_spring_week
    typical_summer = stat.typical_summer_week
    typical_autumn = stat.typical_autumn_week

    assert str(extreme_cold) == '1/27 to 2/2 between 0 and 23 @1'
    assert str(extreme_hot) == '7/13 to 7/19 between 0 and 23 @1'
    assert str(typical_winter) == '12/22 to 12/28 between 0 and 23 @1'
    assert str(typical_spring) == '4/26 to 5/2 between 0 and 23 @1'
    assert str(typical_summer) == '8/24 to 8/30 between 0 and 23 @1'
    assert str(typical_autumn) == '10/27 to 11/2 between 0 and 23 @1'


def test_dict_methods():
    """Test dictionary serialization methods."""
    relative_path = './tests/fixtures/stat/chicago.stat'
    stat_obj = STAT(relative_path)

    stat_dict = stat_obj.to_dict()
    rebuilt_stat = STAT.from_dict(stat_dict)
    assert stat_dict == rebuilt_stat.to_dict()



# coding=utf-8
"""Generic data type."""
from __future__ import division

from .base import DataTypeBase


class RValue(DataTypeBase):
    """R Value
    """
    _units = ('m2-K/W', 'h-ft2-F/Btu', 'clo')
    _si_units = ('m2-K/W', 'clo')
    _ip_units = ('h-ft2-F/Btu', 'clo')
    _min = 0
    _abbreviation = 'Rval'

    def _m2K_W_to_hft2F_Btu(self, value):
        return value * 5.678263337

    def _m2K_W_to_clo(self, value):
        return value / 0.155

    def _hft2F_Btu_to_m2K_W(self, value):
        return value / 5.678263337

    def _clo_to_m2K_W(self, value):
        return value * 0.155

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('m2-K/W', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'h-ft2-F/Btu', from_unit), 'h-ft2-F/Btu'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'm2-K/W', from_unit), 'm2-K/W'


class ClothingInsulation(RValue):
    _abbreviation = 'Rclo'
    _unit_descr = {0: 'No Clothing', 0.5: 'T-shirt + Shorts', 1: '3-piece Suit'}



# coding=utf-8
"""Module for visualization of hourly data collections."""
from __future__ import division

from .analysisperiod import AnalysisPeriod
from .datacollection import HourlyContinuousCollection, HourlyDiscontinuousCollection
from .graphic import GraphicContainer
from .legend import LegendParameters

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug_geometry.geometry3d.mesh import Mesh3D


class HourlyPlot(object):
    """Object for visualization of hourly data collections.

    Args:
        data_collection: A HourlyContinuousCollection or HourlyDiscontinuousCollection
            which will be used to generate the hourly plot.
        legend_parameters: An optional LegendParameter object to change the display
            of the HourlyPlot (Default: None).
        base_point: An optional Point3D to be used as a starting point to generate
            the geometry of the plot (Default: (0, 0, 0)).
        x_dim: An optional number to set the X dimension of the mesh cells (Default: 1).
        y_dim: An optional number to set the Y dimension of the mesh cells (Default: 4).
        z_dim: An optional number to set the Z dimension of the entire chart. This
            will be used to make the colored_mesh3d of the chart vary in the Z
            dimension according to the data. The value input here should usually be
            several times larger than the x_dim or y_dim in order to be noticeable
            (e.g. 100). If 0, the colored_mesh3d will simply be flat. (Default: 0).

    Properties:
        * data_collection
        * legend_parameters
        * base_point
        * x_dim
        * y_dim
        * z_dim
        * colored_mesh2d
        * colored_mesh3d
        * legend
        * chart_border2d
        * chart_border3d
        * hour_lines2d
        * hour_lines3d
        * hour_label_points2d
        * hour_label_points3d
        * hour_labels
        * month_lines2d
        * month_lines3d
        * month_label_points2d
        * month_label_points3d
        * month_labels
        * title_text
        * lower_title_location
        * upper_title_location
        * analysis_period
        * values
        * colors
    """
    __slots__ = ('_data_collection', '_base_point', '_x_dim', '_y_dim', '_z_dim',
                 '_num_y', '_num_x', '_container', '_hour_points', '_hour_text',
                 '_month_points', '_month_label_points', '_month_text')

    HOUR_LABELS = (0, 6, 12, 18, 24)

    def __init__(self, data_collection, legend_parameters=None, base_point=Point3D(),
                 x_dim=1, y_dim=4, z_dim=0):
        """Initialize hourly plot."""
        # check the input objects
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        assert isinstance(data_collection, acceptable_colls), 'HourlyPlot data_collection' \
            ' must be a HourlyContinuousCollection or HourlyDiscontinuousCollection. ' \
            'Got {}.'.format(type(data_collection))
        assert isinstance(base_point, Point3D), 'Expected Point3D for ' \
            'HourlyPlot base point. Got {}.'.format(type(base_point))

        # ensure the analysis period of the data collection has been validated
        if not data_collection.validated_a_period:
            data_collection = data_collection.validate_analysis_period()

        # assign the inputs as properties of this data collection
        self._data_collection = data_collection.to_immutable()
        self._base_point = base_point
        self._x_dim = self._check_dim(x_dim, 'x_dim')
        self._y_dim = self._check_dim(y_dim, 'y_dim')
        self._z_dim = 0
        if z_dim != 0:
            self._z_dim = self._check_dim(z_dim, 'z_dim')

        # set properties to None that will be computed later
        self._hour_points = None
        self._hour_text = None
        self._month_points = None
        self._month_label_points = None
        self._month_text = None

        # create the graphic container from the inputs
        a_per = self.analysis_period
        if not a_per.is_reversed:
            self._num_x = a_per.end_time.doy - a_per.st_time.doy + 1
        else:
            num_doys = 365 if not a_per.is_leap_year else 366
            self._num_x = num_doys - a_per.st_time.doy + a_per.end_time.doy + 1
        if a_per.end_hour == 23:  # the last datetime is not included
            self._num_y = ((a_per.end_hour - a_per.st_hour + 1) * a_per.timestep)
        else:  # the last datetime is included
            self._num_y = ((a_per.end_hour - a_per.st_hour) * a_per.timestep) + 1
        x_dist = x_dim * self._num_x
        y_dist = y_dim * self._num_y
        max_pt = Point3D(base_point.x + x_dist, base_point.y + y_dist, base_point.z)
        self._container = GraphicContainer(
            data_collection.values, base_point, max_pt, legend_parameters,
            data_collection.header.data_type, data_collection.header.unit)

    @classmethod
    def from_z_dim_per_unit(cls, data_collection, legend_parameters=None,
                            base_point=Point3D(), x_dim=1, y_dim=4, z_dim_per_unit=0):
        """Create HourlyPlot with an option to set the Z dimension on a per-unit basis.

        This is useful in cases of wanting to compare the Z dimensions of two plots
        to each other.

        Args:
            data_collection: A HourlyContinuousCollection or HourlyDiscontinuousCollection
                which will be used to generate the hourly plot.
            legend_parameters: An optional LegendParameter object to change the display
                of the HourlyPlot (Default: None).
            base_point: An optional Point3D to be used as a starting point to generate
                the geometry of the plot (Default: (0, 0, 0)).
            x_dim: Optional number to set the X dimension of the mesh cells (Default: 1).
            y_dim: Optional number to set the Y dimension of the mesh cells (Default: 4).
            z_dim: Optional number to set the Z dimension per unit of value in the input
                data_collection. If 0, the colored_mesh3d will be flat. (Default: 0).
        """
        z_dim = (data_collection.max - data_collection.min) * z_dim_per_unit
        return cls(data_collection, legend_parameters, base_point, x_dim, y_dim, z_dim)

    @property
    def data_collection(self):
        """The data collection assigned to this hourly plot."""
        return self._data_collection

    @property
    def legend_parameters(self):
        """The legend parameters customizing this hourly plot."""
        return self._container.legend_parameters

    @property
    def base_point(self):
        """Point3D for the base point of this hourly plot."""
        return self._base_point

    @property
    def x_dim(self):
        """A number for the X dimension of each cell of the hourly plot."""
        return self._x_dim

    @property
    def y_dim(self):
        """A number for the Y dimension of each cell of the hourly plot."""
        return self._y_dim

    @property
    def z_dim(self):
        """A number for the Z dimension of the entire hourly plot."""
        return self._z_dim

    @property
    def colored_mesh2d(self):
        """Get a colored Mesh2D for this graphic."""
        return self._compute_colored_mesh2d()

    @property
    def colored_mesh3d(self):
        """Get a colored Mesh3D for this graphic.

        Note that this will be the same as the colored_mesh2d if the z_dim
        value is 0.
        """
        return self._compute_colored_mesh3d()

    @property
    def legend(self):
        """The legend assigned to this graphic."""
        return self._container._legend

    @property
    def chart_border2d(self):
        """Get a Polyline2D for the border of the plot."""
        base_pt = Point2D(self._base_point.x, self._base_point.y)
        width = self._container.max_point.x - self._container.min_point.x
        height = self._container.max_point.y - self._container.min_point.y
        pgon = Polygon2D.from_rectangle(base_pt, Vector2D(0, 1), width, height)
        return Polyline2D.from_polygon(pgon)

    @property
    def chart_border3d(self):
        """Get a Polyline3D for the border of the plot."""
        plane = Plane(o=Point3D(0, 0, self._base_point.z))
        return Polyline3D.from_polyline2d(self.chart_border2d, plane)

    @property
    def hour_lines2d(self):
        """Get a list of LineSegment2D for the 6-hour intervals of the chart."""
        if not self._hour_points:
             self._compute_static_hour_line_pts()
        vec = Vector2D(self._num_x * self._x_dim)
        return [LineSegment2D(pt, vec) for pt in self._hour_points]

    @property
    def hour_lines3d(self):
        """Get a list of LineSegment3D for the 6-hour intervals of the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        vec = Vector3D(self._num_x * self._x_dim)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in self._hour_points]

    @property
    def hour_label_points2d(self):
        """Get a list of Point2Ds for the 6-hour text labels for the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x - txt_hght * 2, pt.y) for pt in self._hour_points]

    @property
    def hour_label_points3d(self):
        """Get a list of Point3Ds for the 6-hour text labels for the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x - txt_hght * 2, pt.y, self._base_point.z)
                for pt in self._hour_points]

    @property
    def hour_labels(self):
        """Get a list of text strings for the 6-hour labels for the chart."""
        if not self._hour_text:
            self._compute_static_hour_line_pts()
        return self._hour_text

    @property
    def month_lines2d(self):
        """Get a list of LineSegment2D for the month intervals of the chart."""
        if not self._month_points:
            self._compute_month_line_pts()
        vec = Vector2D(0, self._num_y * self._y_dim)
        return [LineSegment2D(pt, vec) for pt in self._month_points]

    @property
    def month_lines3d(self):
        """Get a list of LineSegment3D for the month intervals of the chart."""
        if not self._month_points:
            self._compute_month_line_pts()
        vec = Vector3D(0, self._num_y * self._y_dim, 0)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in self._month_points]

    @property
    def month_label_points2d(self):
        """Get a list of Point2Ds for the month text labels for the chart."""
        if not self._month_label_points:
            self._compute_month_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x, pt.y - txt_hght) for pt in self._month_label_points]

    @property
    def month_label_points3d(self):
        """Get a list of Point3Ds for the month text labels for the chart."""
        if not self._month_label_points:
            self._compute_month_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x, pt.y - txt_hght, self._base_point.z)
                for pt in self._month_label_points]

    @property
    def month_labels(self):
        """Get a list of text strings for the month labels for the chart."""
        if not self._month_text:
            self._compute_month_line_pts()
        return self._month_text

    @property
    def title_text(self):
        """Text string for the title of the hourly plot."""
        title_array = ['{} ({})'.format(self._data_collection.header.data_type,
                                        self._data_collection.header.unit)]
        title_array.append(str(self._data_collection.header.analysis_period))
        for key, val in self._data_collection.header.metadata.items():
            title_array.append('{}: {}'.format(key, val))
        return '\n'.join(title_array)

    @property
    def lower_title_location(self):
        """A Plane for the lower location of title text."""
        pln = self._container.lower_title_location
        txt_hght = self.legend_parameters.text_height
        return Plane(pln.n, Point3D(pln.o.x, pln.o.y - txt_hght * 2, pln.o.z))

    @property
    def upper_title_location(self):
        """A Plane for the upper location of title text."""
        num_lines = self.title_text.count('\n') + 4
        pln = self._container.upper_title_location
        txt_hght = self.legend_parameters.text_height
        return Plane(pln.n, Point3D(pln.o.x, pln.o.y + txt_hght * num_lines, pln.o.z))

    @property
    def analysis_period(self):
        """The AnalysisPeriod assigned to the hourly plot's data collection."""
        return self._data_collection.header.analysis_period

    @property
    def values(self):
        """A list of values assigned to this hourly plot."""
        return self._data_collection.values

    @property
    def colors(self):
        """A list of colors assigned to the mesh faces of this hourly plot."""
        return self._container.value_colors

    def custom_hour_lines2d(self, hour_labels):
        """Get a list of LineSegment2D for a list of numbers representing hour labels.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text = self._compute_hour_line_pts(hour_labels)
        vec = Vector2D(self._num_x * self._x_dim)
        return [LineSegment2D(pt, vec) for pt in _hour_points]

    def custom_hour_lines3d(self, hour_labels):
        """Get a list of LineSegment3D for the 6-hour intervals of the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text = self._compute_hour_line_pts(hour_labels)
        vec = Vector3D(self._num_x * self._x_dim)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in _hour_points]

    def custom_hour_label_points2d(self, hour_labels):
        """Get a list of Point2Ds for the 6-hour text labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text = self._compute_hour_line_pts(hour_labels)
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x - txt_hght * 2, pt.y) for pt in _hour_points]

    def custom_hour_label_points3d(self, hour_labels):
        """Get a list of Point3Ds for the 6-hour text labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text = self._compute_hour_line_pts(hour_labels)
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x - txt_hght * 2, pt.y, self._base_point.z)
                for pt in _hour_points]

    def custom_hour_labels(self, hour_labels):
        """Get a list of text strings for the 6-hour labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text = self._compute_hour_line_pts(hour_labels)
        return _hour_text

    def _compute_colored_mesh2d(self):
        """Compute a colored mesh from this object's data collection."""
        # generate the base mesh as a stanadard grid
        _colored_mesh2d = Mesh2D.from_grid(
            self.base_point, self._num_x, self._num_y, self.x_dim, self.y_dim)

        # remove any faces in the base mesh that do not represent the data
        if not isinstance(self.data_collection, HourlyContinuousCollection):
            # get a pattern of booleans for whether mesh faces should be included
            data_coll_moys = [dt.moy for dt in self.data_collection.datetimes]
            data_coll_moys.append(527100)  # extra value for the end of the list
            found_i = 0
            mesh_pattern = []
            for moy in self.analysis_period.moys:
                if moy == data_coll_moys[found_i]:
                    mesh_pattern.append(True)
                    found_i += 1
                else:
                    mesh_pattern.append(False)
            _colored_mesh2d = _colored_mesh2d.remove_faces_only(mesh_pattern)

        # assign the colors to the mesh
        _colored_mesh2d.colors = self.colors
        return _colored_mesh2d

    def _compute_colored_mesh3d(self):
        """Compute a colored mesh from this object's data collection."""
        _colored_mesh3d = Mesh3D.from_mesh2d(
            self.colored_mesh2d, Plane(o=Point3D(0, 0, self._container.min_point.z)))
        if self.z_dim != 0:
            _colored_mesh3d = _colored_mesh3d.height_field_mesh(
                self.data_collection.values, (0, self.z_dim))
        return _colored_mesh3d

    def _compute_static_hour_line_pts(self):
        """Compute the points for the hour lines and labels."""
        self._hour_points, self._hour_text = \
                 self._compute_hour_line_pts(self.HOUR_LABELS)

    def _compute_hour_line_pts(self, hour_labels):
        """Compute the points for the hour lines and labels."""
        st_hr = self.analysis_period.st_hour
        end_hr = self.analysis_period.end_hour + 1
        t_step = self.analysis_period.timestep
        _hour_points = []
        _hour_text = []
        last_hr = 0
        for hr in hour_labels:
            if st_hr <= hr <= end_hr:
                pt_y = self.base_point.y + (hr - st_hr) * self.y_dim * t_step
                pt = Point2D(self._container.min_point.x, pt_y)
                _hour_points.append(pt)
                hr_val = hr if hr <= 12 else hr - 12
                am_pm = 'PM' if 12 <= hr < 24 else 'AM'
                if hr_val == 0:
                    hr_val = 12
                _hour_text.append('{} {}'.format(hr_val, am_pm))
                last_hr = hr
        if self.analysis_period.timestep > 1 and last_hr == end_hr:
            _hour_points.pop(-1)
            _hour_text.pop(-1)
        return _hour_points, _hour_text

    def _compute_month_line_pts(self):
        """Compute the points for the hour lines and labels."""
        # extract several reused properties from the analysis period
        st_mon = self.analysis_period.st_month
        end_mon = self.analysis_period.end_month
        st_day = self.analysis_period.st_day
        end_day = self.analysis_period.end_day
        dpm = AnalysisPeriod.NUMOFDAYSEACHMONTHLEAP if self.analysis_period.is_leap_year \
            else AnalysisPeriod.NUMOFDAYSEACHMONTH

        # create a list of days in each of the months and collect month text
        self._month_text = [AnalysisPeriod.MONTHNAMES[st_mon]]
        days_list = []
        if st_mon == end_mon:
            days_list = [dpm[st_mon - 1] + 1 - st_day + (end_day - dpm[st_mon - 1])]
        else:
            days_list.append(dpm[st_mon - 1] + 1 - st_day)
        month_count = len(self.analysis_period.months_int) - 2
        for count, month in enumerate(self.analysis_period.months_int[1:]):
            self._month_text.append(AnalysisPeriod.MONTHNAMES[month])
            if count == month_count:
                days_list.append(end_day)
            else:
                days_list.append(dpm[month - 1])

        # create the points for each of the month labels
        b_pt = self._base_point
        self._month_label_points = \
            [Point2D(b_pt.x + (days_list[0] / 2) * self.x_dim, b_pt.y)]
        for mon_days in days_list[1:]:
            prev_pt = self._month_label_points[-1]
            new_pt = Point2D(prev_pt.x + mon_days * self.x_dim, prev_pt.y)
            self._month_label_points.append(new_pt)

        # create the points for each of the month lines
        self._month_points = []
        if len(days_list) > 1:
            prev_x = b_pt.x + days_list[0] * self.x_dim
            for mon_days in days_list[1:]:
                self._month_points.append(Point2D(prev_x, b_pt.y))
                prev_x += mon_days * self.x_dim

    @staticmethod
    def _check_dim(dim_value, dim_name):
        """Check a given value for a dimension input."""
        assert isinstance(dim_value, (float, int)), 'Expected number for ' \
            'HourlyPlot {}. Got {}.'.format(dim_name, type(dim_value))
        assert dim_value > 0, 'HourlyPlot {} must be greater than 0. ' \
            'Got {}.'.format(dim_name, dim_value)
        return dim_value

    def __repr__(self):
        """Hourly Plot representation."""
        return 'Hourly Plot:\n{}'.format(self.data_collection.header)



# coding=utf-8
from ladybug.rootfinding import secant, bisect


def test_secant():
    """Test the secant rootfinding method."""
    def funct(x):
        return (x + 1) ** 2 - 1
    
    root_val = secant(-5, 5, funct, 0.001)
    assert root_val < 1e-3


def test_bisect():
    """Test the bisect rootfinding method."""
    def funct(x):
        return (x + 1) ** 2 - 1
    
    root_val = bisect(-5, 5, funct, 0.001, 0)
    assert root_val < 1e-3



"""Ladybug configurations.

Import this into every module where access configurations are needed.

Usage:
    from ladybug.config import folders
    print(folders.default_epw_folder)
    folders.default_epw_folder = "C:/epw_data"
"""
import os
import sys
import json


class Folders(object):
    """Ladybug folders.

    Args:
        config_file: The path to the config.json file from which folders are loaded.
            If None, the config.json module included in this package will be used.
            Default: None.
        mute: If False, the paths to the various folders will be printed as they
            are found. If True, no printing will occur upon initialization of this
            class. Default: True.

    Properties:
        * ladybug_tools_folder
        * default_epw_folder
        * config_file
        * mute
    """

    def __init__(self, config_file=None, mute=True):
        # set the mute value
        self.mute = bool(mute)

        # load paths from the config JSON file 
        self.config_file  = config_file

    @property
    def ladybug_tools_folder(self):
        """Get or set the path to the ladybug tools installation folder."""
        return self._ladybug_tools_folder
    
    @ladybug_tools_folder.setter
    def ladybug_tools_folder(self, path):
        if not path:  # check the default location for epw files
            path = self._find_default_ladybug_tools_folder()

        self._ladybug_tools_folder = path

        if not self.mute and self._ladybug_tools_folder:
            print('Path to the ladybug tools installation folder is set to: '
                  '{}'.format(self._ladybug_tools_folder))

    @property
    def default_epw_folder(self):
        """Get or set the path to the default folder where EPW files are stored."""
        return self._default_epw_folder
    
    @default_epw_folder.setter
    def default_epw_folder(self, path):
        if not path:  # check the default location for epw files
            path = self._find_default_epw_folder()

        self._default_epw_folder = path

        if not self.mute and self._default_epw_folder:
            print('Path to the default epw folder is set to: '
                  '{}'.format(self._default_epw_folder))
    
    @property 
    def config_file(self):
        """Get or set the path to the config.json file from which folders are loaded.
        
        Setting this to None will result in using the config.json module included
        in this package.
        """
        return self._config_file

    @config_file.setter
    def config_file(self, cfg):
        if cfg is None:
            cfg = os.path.join(os.path.dirname(__file__), 'config.json')
        self._load_from_file(cfg)
        self._config_file = cfg

    def _load_from_file(self, file_path):
        """Set all of the the properties of this object from a config JSON file.
        
        Args:
            file_path: Path to a JSON file containing the file paths. A sample of this
                JSON is the config.json file within this package.
        """
        # check the default file path
        assert os.path.isfile(file_path), \
            ValueError('No file found at {}'.format(file_path))

        # set the default paths to be all blank
        default_path = {
            "ladybug_tools_folder": r'',
            "default_epw_folder": r''
        }

        with open(file_path, 'r') as cfg:
            try:
                paths = json.load(cfg)
            except Exception as e:
                print('Failed to load paths from {}.\nThey will be set to defaults '
                      'instead\n{}'.format(file_path, e))
            else:
                for key, p in paths.items():
                    if not key.startswith('__') and p.strip():
                        default_path[key] = p.strip()

        # set paths for the ladybug_tools_folder and default_epw_folder
        self.ladybug_tools_folder = default_path["ladybug_tools_folder"]
        self.default_epw_folder = default_path["default_epw_folder"]

    def _find_default_epw_folder(self):
        """Find the the default EPW folder in its usual location.
        
        An attempt will be made to create the directory if it does not already exist.
        """
        epw_folder = os.path.join(self.ladybug_tools_folder, 'resources', 'weather')
        if not os.path.isdir(epw_folder):
            try:
                os.makedirs(epw_folder)
            except Exception as e:
                raise OSError('Failed to create default epw '
                              'folder: %s\n%s' % (epw_folder, e))
        return epw_folder

    @staticmethod
    def _find_default_ladybug_tools_folder():
        """Find the the default ladybug_tools folder in its usual location.
        
        An attempt will be made to create the directory if it does not already exist.
        """
        home_folder = os.getenv('HOME') or os.path.expanduser('~')
        install_folder = os.path.join(home_folder, 'ladybug_tools')
        if not os.path.isdir(install_folder):
            try:
                os.makedirs(install_folder)
            except Exception as e:
                raise OSError('Failed to create default ladybug tools installation '
                              'folder: %s\n%s' % (install_folder, e))
        return install_folder


"""Object possesing all key folders within the configuration."""
folders = Folders()



# coding=utf-8
from ladybug.wea import Wea
from ladybug.location import Location
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.epw import EPW

import pytest
import os


def test_from_file():
    """Test import from wea file."""
    wea_file = './tests/fixtures/wea/san_francisco_10min.wea'
    with pytest.raises(Exception):
        Wea.from_file(wea_file)  # wrong timestep

    wea = Wea.from_file(wea_file, 6)
    assert wea.direct_normal_irradiance[45] == 88
    assert wea.diffuse_horizontal_irradiance[45] == 1
    assert wea.direct_normal_irradiance[46] == 313
    assert wea.diffuse_horizontal_irradiance[46] == 3


def test_from_epw():
    """Test import from epw"""
    epw_path = './tests/fixtures/epw/chicago.epw'
    wea_from_epw = Wea.from_epw_file(epw_path)

    assert wea_from_epw.location.city == 'Chicago Ohare Intl Ap'
    assert wea_from_epw.timestep == 1
    assert wea_from_epw.direct_normal_irradiance[7] == 22
    assert wea_from_epw.direct_normal_irradiance.datetimes[7].hour == 7
    assert wea_from_epw.direct_normal_irradiance.datetimes[7].minute == 0
    assert wea_from_epw.direct_normal_irradiance[8] == 397
    assert wea_from_epw.direct_normal_irradiance.datetimes[8].hour == 8
    assert wea_from_epw.direct_normal_irradiance.datetimes[8].minute == 0
    # diffuse horizontal irradiance
    assert wea_from_epw.diffuse_horizontal_irradiance[7] == 10
    assert wea_from_epw.diffuse_horizontal_irradiance.datetimes[7].hour == 7
    assert wea_from_epw.diffuse_horizontal_irradiance.datetimes[7].minute == 0
    assert wea_from_epw.diffuse_horizontal_irradiance[8] == 47
    assert wea_from_epw.diffuse_horizontal_irradiance.datetimes[8].hour == 8
    assert wea_from_epw.diffuse_horizontal_irradiance.datetimes[8].minute == 0


def test_equality():
    wea_file = './tests/fixtures/wea/chicago.wea'
    wea_1 = Wea.from_file(wea_file)
    wea_2 = Wea.from_file(wea_file)
    assert wea_1 == wea_2


def test_from_stat():
    """Test import from stat"""
    stat_path = './tests/fixtures/stat/chicago.stat'
    wea_from_stat = Wea.from_stat_file(stat_path)

    assert wea_from_stat.location.city == 'Chicago Ohare Intl Ap'
    assert wea_from_stat.timestep == 1
    assert wea_from_stat.diffuse_horizontal_irradiance[0] == \
        pytest.approx(0, rel=1e-3)
    assert wea_from_stat.direct_normal_irradiance[0] == \
        pytest.approx(0, rel=1e-3)
    assert wea_from_stat.diffuse_horizontal_irradiance[12] == \
        pytest.approx(87.44171, rel=1e-3)
    assert wea_from_stat.direct_normal_irradiance[12] == \
        pytest.approx(810.693919, rel=1e-3)


def test_from_stat_missing_optical():
    """Test import from a stat file that is missing optical data"""
    stat_path = './tests/fixtures/stat/santamonica.stat'
    with pytest.raises(ValueError,
                       match='Stat file contains no optical data.'):
        Wea.from_stat_file(stat_path)


def test_from_clear_sky():
    """Test from original clear sky"""
    location = Location(
        'Chicago Ohare Intl Ap', '-', 'USA', 41.98, -87.92, -6.0, 201.0)
    wea_from_clear_sky = Wea.from_ashrae_clear_sky(location)

    assert wea_from_clear_sky.location.city == 'Chicago Ohare Intl Ap'
    assert wea_from_clear_sky.timestep == 1
    assert wea_from_clear_sky.diffuse_horizontal_irradiance[0] == \
        pytest.approx(0, rel=1e-3)
    assert wea_from_clear_sky.direct_normal_irradiance[0] == \
        pytest.approx(0, rel=1e-3)
    assert wea_from_clear_sky.diffuse_horizontal_irradiance[12] == \
        pytest.approx(60.72258, rel=1e-3)
    assert wea_from_clear_sky.direct_normal_irradiance[12] == \
        pytest.approx(857.00439, rel=1e-3)


def test_from_zhang_huang():
    """Test from zhang huang solar model"""
    path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(path)

    wea_from_zh = Wea.from_zhang_huang_solar(epw.location,
                                             epw.total_sky_cover.values,
                                             epw.relative_humidity.values,
                                             epw.dry_bulb_temperature.values,
                                             epw.wind_speed.values,
                                             epw.atmospheric_station_pressure.values)

    assert wea_from_zh.location.city == 'Chicago Ohare Intl Ap'
    assert wea_from_zh.timestep == 1
    assert wea_from_zh.global_horizontal_irradiance[0] == \
        pytest.approx(0, rel=1e-1)
    assert wea_from_zh.global_horizontal_irradiance[12] == \
        pytest.approx(417.312, rel=1e-1)
    assert wea_from_zh.direct_normal_irradiance[12] == \
        pytest.approx(654.52, rel=1e-1)
    assert wea_from_zh.diffuse_horizontal_irradiance[12] == \
        pytest.approx(144.51, rel=1e-1)


def test_zhang_huang_accuracy():
    """Test zhang huang solar model to ensure that average error is within
    25% of actual solar."""
    path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(path)

    wea = Wea.from_zhang_huang_solar(epw.location,
                                     epw.total_sky_cover.values,
                                     epw.relative_humidity.values,
                                     epw.dry_bulb_temperature.values,
                                     epw.wind_speed.values,
                                     epw.atmospheric_station_pressure.values)
    # test global horizontal radiation
    glob_horiz_error = [abs(i - j) for i, j in zip(
        epw.global_horizontal_radiation,
        wea.global_horizontal_irradiance)]
    avg_glob_horiz_error = sum(glob_horiz_error) / sum(
        epw.global_horizontal_radiation)
    assert (sum(glob_horiz_error) / 8760) < 50
    assert avg_glob_horiz_error < 0.25

    # test direct normal radiation
    dir_normal_error = [abs(i - j) for i, j in zip(
        epw.direct_normal_radiation, wea.direct_normal_irradiance)]
    avg_dir_normal_error = sum(dir_normal_error) / sum(
        epw.direct_normal_radiation)
    assert sum(dir_normal_error) / 8760 < 100
    assert avg_dir_normal_error < 0.5

    # test diffuse horizontal radiation
    dif_horiz_error = [abs(i - j) for i, j in zip(
        epw.diffuse_horizontal_radiation,
        wea.diffuse_horizontal_irradiance)]
    avg_dif_horiz_error = sum(dif_horiz_error) / sum(
        epw.diffuse_horizontal_radiation)
    assert sum(dif_horiz_error) / 8760 < 50
    assert avg_dif_horiz_error < 0.5


def test_dict_methods():
    """Test JSON serialization methods"""
    epw_path = './tests/fixtures/epw/chicago.epw'
    wea = Wea.from_epw_file(epw_path)

    assert wea.to_dict() == Wea.from_dict(wea.to_dict()).to_dict()


def test_import_epw():
    """Test to compare import from epw with its dict version."""
    epw_path = './tests/fixtures/epw/chicago.epw'

    wea_from_epw = Wea.from_epw_file(epw_path)

    wea_dict = wea_from_epw.to_dict()
    wea_from_dict = Wea.from_dict(wea_dict)
    assert wea_from_dict.direct_normal_irradiance.values == \
        wea_from_epw.direct_normal_irradiance.values
    assert wea_from_dict.diffuse_horizontal_irradiance.values == \
        wea_from_epw.diffuse_horizontal_irradiance.values


def test_import_stat():
    """Test to compare import from stat with its dict version."""
    stat_path = './tests/fixtures/stat/chicago.stat'
    wea_from_stat = Wea.from_stat_file(stat_path)

    wea_dict = wea_from_stat.to_dict()
    wea_from_dict = Wea.from_dict(wea_dict)
    assert wea_from_dict.direct_normal_irradiance.values == \
        wea_from_stat.direct_normal_irradiance.values
    assert wea_from_dict.diffuse_horizontal_irradiance.values == \
        wea_from_stat.diffuse_horizontal_irradiance.values


def test_write_wea():
    """Test the write Wea file capability."""
    stat_path = './tests/fixtures/stat/chicago.stat'
    wea_from_stat = Wea.from_stat_file(stat_path)

    wea_path = './tests/fixtures/wea/chicago_stat.wea'
    hrs_path = './tests/fixtures/wea/chicago_stat.hrs'
    hoys = range(8760)
    wea_from_stat.write(wea_path, hoys, True)

    assert os.path.isfile(wea_path)
    assert os.stat(wea_path).st_size > 1
    assert os.path.isfile(hrs_path)
    assert os.stat(hrs_path).st_size > 1

    # check the order of the data in the file
    with open(wea_path) as wea_f:
        lines = wea_f.readlines()
        assert float(lines[6].split(' ')[-2]) == \
            pytest.approx(
                wea_from_stat.direct_normal_irradiance[0], rel=1e-1)
        assert int(lines[6].split(' ')[-1]) == \
            wea_from_stat.diffuse_horizontal_irradiance[0]
        assert float(lines[17].split(' ')[-2]) == \
            pytest.approx(
                wea_from_stat.direct_normal_irradiance[11], rel=1e-1)
        assert float(lines[17].split(' ')[-1]) == \
            pytest.approx(
                wea_from_stat.diffuse_horizontal_irradiance[11], rel=1e-1)

    os.remove(wea_path)
    os.remove(hrs_path)


def test_global_and_direct_horizontal():
    """Test the global horizontal irradiance on method."""
    stat_path = './tests/fixtures/stat/chicago.stat'
    wea_from_stat = Wea.from_stat_file(stat_path)

    diffuse_horiz_rad = wea_from_stat.diffuse_horizontal_irradiance
    direct_horiz_rad = wea_from_stat.direct_horizontal_irradiance
    glob_horiz_rad = wea_from_stat.global_horizontal_irradiance

    assert [x for x in glob_horiz_rad] == pytest.approx(
        [x + y for x, y in zip(diffuse_horiz_rad, direct_horiz_rad)], rel=1e-3)


def test_directional_irradiance():
    """Test the directional irradiance method."""
    stat_path = './tests/fixtures/stat/chicago.stat'
    wea_from_stat = Wea.from_stat_file(stat_path)

    srf_total, srf_direct, srf_diffuse, srf_reflect = \
        wea_from_stat.directional_irradiance(90)
    diffuse_horiz_rad = wea_from_stat.diffuse_horizontal_irradiance
    direct_horiz_rad = wea_from_stat.direct_horizontal_irradiance
    glob_horiz_rad = wea_from_stat.global_horizontal_irradiance

    assert srf_total.values == pytest.approx(glob_horiz_rad.values, rel=1e-3)
    assert srf_direct.values == pytest.approx(direct_horiz_rad.values, rel=1e-3)
    assert srf_diffuse.values == pytest.approx(diffuse_horiz_rad.values, rel=1e-3)
    assert srf_reflect.values == pytest.approx([0] * 8760, rel=1e-3)


def test_estimate_illuminance():
    """Test the directional irradiance method."""
    epw_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(epw_path)
    wea = Wea.from_epw_file(epw_path)

    glob_ill, dir_ill, diff_ill, zen_lum = \
        wea.estimate_illuminance_components(epw.dew_point_temperature)

    assert glob_ill.bounds[0] == pytest.approx(0, rel=1e-3)
    assert 100000 < glob_ill.bounds[1] < 125000
    assert dir_ill.bounds[0] == pytest.approx(0, rel=1e-3)
    assert 75000 < dir_ill.bounds[1] < 125000
    assert diff_ill.bounds[0] == pytest.approx(0, rel=1e-3)
    assert 50000 < diff_ill.bounds[1] < 100000
    assert zen_lum.bounds[0] == pytest.approx(0, rel=1e-3)
    assert zen_lum.bounds[1] < 35000


def test_leap_year():
    """Test clear sky with leap year."""
    location = Location(
        'Chicago Ohare Intl Ap', '-', 'USA', 41.98, -87.92, -6.0, 201.0)
    wea = Wea.from_ashrae_clear_sky(location, is_leap_year=True)

    assert wea.diffuse_horizontal_irradiance.datetimes[1416].month == 2
    assert wea.diffuse_horizontal_irradiance.datetimes[1416].day == 29
    assert wea.diffuse_horizontal_irradiance.datetimes[1416].hour == 0

    assert wea.diffuse_horizontal_irradiance.datetimes[1416 + 12].month == 2
    assert wea.diffuse_horizontal_irradiance.datetimes[1416 + 12].day == 29
    assert wea.diffuse_horizontal_irradiance.datetimes[1416 + 12].hour == 12



# These must be imported here in this order because of how they reference eachother.
import ladybug._datacollectionbase
import ladybug.datacollection
import ladybug.datacollectionimmutable

from ladybug.logutil import get_logger

import sys
import importlib
import pkgutil

# set up the logger
logger = get_logger(__name__)

#  find and import ladybug extensions
#  this is a critical step to add additional functionalities to ladybug core library.
extensions = {}
for finder, name, ispkg in pkgutil.iter_modules():
    if not name.startswith('ladybug_') or name.count('_') > 1:
        continue
    try:
        extensions[name] = importlib.import_module(name)
    except Exception:
        logger.exception('Failed to import {0}!'.format(name))
    else:
        logger.info('Successfully imported Ladybug plugin: {}'.format(name))



# coding=utf-8
from ladybug.compass import Compass

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.arc import Arc2D
from ladybug_geometry.geometry3d.pointvector import Point3D


def test_init_compass():
    """Test the initialization of Compass and basic properties."""
    compass = Compass()

    str(compass)  # test the string representation
    hash(compass)  # test to be sure the compass is hash-able
    assert compass.radius == 100
    assert compass.center == Point2D()
    assert compass.north_angle == 0
    assert compass.north_vector == Vector2D(0, 1)
    assert compass.spacing_factor == 0.15
    assert compass.min_point.is_equivalent(Point2D(-115.0, -115.0), 0.01)
    assert compass.max_point.is_equivalent(Point2D(115.0, 115.0), 0.01)
    assert isinstance(compass.inner_boundary_circle, Arc2D)
    assert len(compass.all_boundary_circles) == 3
    for circ in compass.all_boundary_circles:
        assert isinstance(circ, Arc2D)
    assert len(compass.major_azimuth_points) == len(compass.MAJOR_AZIMUTHS)
    for pt in compass.major_azimuth_points:
        assert isinstance(pt, Point2D)
    assert len(compass.major_azimuth_ticks) == len(compass.MAJOR_AZIMUTHS)
    for lin in compass.major_azimuth_ticks:
        assert isinstance(lin, LineSegment2D)
    assert len(compass.minor_azimuth_points) == len(compass.MINOR_AZIMUTHS)
    for pt in compass.minor_azimuth_points:
        assert isinstance(pt, Point2D)
    assert len(compass.minor_azimuth_ticks) == len(compass.MINOR_AZIMUTHS)
    for lin in compass.minor_azimuth_ticks:
        assert isinstance(lin, LineSegment2D)
    assert isinstance(compass.min_point3d(), Point3D)
    assert isinstance(compass.max_point3d(), Point3D)


def test_compass_orthographic():
    """Test the orthographic properties of the Compass."""
    compass = Compass()

    assert len(compass.orthographic_altitude_points) == len(compass.ALTITUDES)
    for pt in compass.orthographic_altitude_points:
        assert isinstance(pt, Point2D)
    assert len(compass.orthographic_altitude_circles) == len(compass.ALTITUDES)
    for lin in compass.orthographic_altitude_circles:
        assert isinstance(lin, Arc2D)


def test_compass_stereographic():
    """Test the stereographic properties of the Compass."""
    compass = Compass()

    assert len(compass.stereographic_altitude_points) == len(compass.ALTITUDES)
    for pt in compass.stereographic_altitude_points:
        assert isinstance(pt, Point2D)
    assert len(compass.stereographic_altitude_circles) == len(compass.ALTITUDES)
    for lin in compass.stereographic_altitude_circles:
        assert isinstance(lin, Arc2D)


def test_set_properties():
    """Test the initialization of Compass and basic properties."""
    compass = Compass()

    compass.radius = 200
    assert compass.radius == 200
    compass.center = Point2D(200, 0)
    assert compass.center == Point2D(200, 0)
    compass.north_angle = 10
    assert compass.north_angle == 10
    assert compass.north_vector != Vector2D(0, 1)
    compass.spacing_factor = 0.2
    assert compass.spacing_factor == 0.2
    assert compass.min_point.is_equivalent(Point2D(-40.0, -240.0), 0.01)
    assert compass.max_point.is_equivalent(Point2D(440.0, 240.0), 0.01)


def test_equality():
    """Test the equality and duplicatiion of of Compass objects."""
    compass = Compass(radius=100)
    compass_dup = compass.duplicate()
    compass_alt = Compass(radius=200)

    assert compass is compass
    assert compass is not compass_dup
    assert compass == compass_dup
    assert compass != compass_alt


def test_label_points_from_angles():
    """Test the label_points_from_angles method."""
    compass = Compass()
    angles = list(range(0, 360, 30))

    pts = compass.label_points_from_angles(angles)
    assert len(pts) == len(angles)
    for pt in pts:
        assert isinstance(pt, Point2D)

    lines = compass.ticks_from_angles(angles)
    assert len(lines) == len(angles)
    for lin in lines:
        assert isinstance(lin, LineSegment2D)



# coding=utf-8
"""Area data type."""
from __future__ import division

from .base import DataTypeBase


class Area(DataTypeBase):
    """Area
    """
    _units = ('m2', 'ft2', 'mm2', 'in2', 'km2', 'mi2', 'cm2', 'ha', 'acre')
    _si_units = ('m2', 'mm2', 'km2', 'cm2', 'ha')
    _ip_units = ('ft2', 'in2', 'mi2', 'acre')
    _min = 0
    _abbreviation = 'A'

    def _m2_to_ft2(self, value):
        return value * 10.7639

    def _m2_to_mm2(self, value):
        return value * 1000000.

    def _m2_to_in2(self, value):
        return value * 1550.

    def _m2_to_km2(self, value):
        return value / 1000000.

    def _m2_to_mi2(self, value):
        return value / 2590000.

    def _m2_to_cm2(self, value):
        return value * 10000.

    def _m2_to_ha(self, value):
        return value / 10000.

    def _m2_to_acre(self, value):
        return value / 4046.86

    def _ft2_to_m2(self, value):
        return value / 10.7639

    def _mm2_to_m2(self, value):
        return value / 1000000.

    def _in2_to_m2(self, value):
        return value / 1550.

    def _km2_to_m2(self, value):
        return value * 1000000.

    def _mi2_to_m2(self, value):
        return value * 2590000.

    def _cm2_to_m2(self, value):
        return value / 10000.

    def _ha_to_m2(self, value):
        return value * 10000.

    def _acre_to_m2(self, value):
        return value * 4046.86

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('m2', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'mm2' or from_unit == 'cm2':
            return self.to_unit(values, 'in2', from_unit), 'in2'
        elif from_unit == 'km2':
            return self.to_unit(values, 'mi2', from_unit), 'mi2'
        elif from_unit == 'ha':
            return self.to_unit(values, 'acre', from_unit), 'acre'
        else:
            return self.to_unit(values, 'ft2', from_unit), 'ft2'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'in2':
            return self.to_unit(values, 'mm2', from_unit), 'mm2'
        elif from_unit == 'mi2':
            return self.to_unit(values, 'km2', from_unit), 'km2'
        elif from_unit == 'acre':
            return self.to_unit(values, 'ha', from_unit), 'ha'
        else:
            return self.to_unit(values, 'm2', from_unit), 'm2'



# coding=utf-8
from __future__ import division

from .location import Location
from .designday import DesignDay
from .futil import write_to_file

import os
import re
import platform
import codecs


class DDY(object):
    """A DDY object containing all of the data of a .ddy file.

    Args:
        location: A Ladybug location object
        design_days: A list of the design days in the ddy file.

    Properties:
        * file_path
        * location
        * design_days
    """
    __slots__ = ('_location', '_design_days', '_file_path')

    def __init__(self, location, design_days):
        assert isinstance(location, Location), 'Expected' \
            ' Location type. Got {}'.format(type(location))

        self._location = location
        self.design_days = design_days
        self._file_path = None

    @classmethod
    def from_dict(cls, data):
        """Create a DDY from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "DDY",
            "location": {},  # ladybug Location schema
            "design_days": []  # list of ladybug DesignDay schemas
            }
        """
        required_keys = ('location', 'design_days')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)

        return cls(Location.from_dict(data['location']),
                   [DesignDay.from_dict(des_day) for des_day in data['design_days']])

    @classmethod
    def from_ddy_file(cls, file_path):
        """Initialize from a ddy file object from an existing ddy file.

        Args:
            file_path: A string representing a complete path to the .ddy file.
        """
        # check that the file is there
        if not os.path.isfile(file_path):
            raise ValueError(
                'Cannot find a .ddy file at {}'.format(file_path))
        if not file_path.lower().endswith('.ddy'):
            raise ValueError(
                'DDY file does not have a .ddy extension.')

        # check the python version and open the file
        if platform.python_implementation() == 'IronPython':
            ddywin = codecs.open(file_path, 'r')
        else:
            ddywin = codecs.open(file_path, 'r', encoding='utf-8', errors='ignore')

        # extract all location and design day definitions from the file
        loc_p = re.compile(r"(Site:Location,(.|\n)*?((;\s*!)|(;\s*\n)|(;\n)))")
        dday_p = re.compile(r"(SizingPeriod:DesignDay,(.|\n)*?((;\s*!)|(;\s*\n)|(;\n)))")
        try:
            ddytxt = ddywin.read()
            loc_matches = loc_p.findall(ddytxt)
            dday_matches = dday_p.findall(ddytxt)
        except Exception as e:  # the file likely doesn't exist
            import traceback
            raise Exception('{}\n{}'.format(e, traceback.format_exc()))
        else: 
            # check to be sure a location and a design day was found
            assert len(loc_matches) > 0, 'No location objects found in .ddy file.'
            assert len(dday_matches) > 0, 'No design day objects found in .ddy file.'

            # build design day and location objects
            location = Location.from_idf(loc_matches[0][0])
            ddays = [DesignDay.from_idf(match[0], location) for match in dday_matches]
        finally:
            ddywin.close()

        cls_ = cls(location, ddays)
        cls_._file_path = os.path.normpath(file_path)
        return cls_

    @classmethod
    def from_design_day(cls, design_day):
        """Initialize from a ddy file object from a ladybug design day object.

        Args:
            design_day: A Ladybug DesignDay object.
        """
        return cls(design_day.location, [design_day])

    def save(self, file_path):
        """Save ddy object as a .ddy file.

        Args:
            file_path: A string representing the path to write the ddy file to.
        """
        data = self.location.to_idf() + '\n\n'
        for d_day in self.design_days:
            data = data + d_day.to_idf() + '\n\n'
        write_to_file(file_path, data, True)

    def filter_by_keyword(self, keyword):
        """Return a list of ddys that have a certain keyword in their name.

        This is useful for selecting out design days from a ddy file that are
        for a specific type of condition (for example, .4% cooling design days)
        """
        filtered_days = []
        for des_day in self._design_days:
            if keyword in des_day.name:
                filtered_days.append(des_day)
        return filtered_days

    @property
    def file_path(self):
        """Get the original .ddy file path.
        
        Will be None if the DDY did not originate from a file.
        """
        return self._file_path

    @property
    def location(self):
        """Get or set the location."""
        return self._location

    @location.setter
    def location(self, data):
        assert isinstance(data, Location), 'Expected' \
            ' Ladybug Location. Got {}'.format(type(data))
        self._location = data
        for dd in self._design_days:
            if dd.location != self._location:
                dd.location = self._location
                print('Updating location of {} to {}.'.format(dd, self._location))

    @property
    def design_days(self):
        """Get or set the design_days."""
        return tuple(self._design_days)

    @design_days.setter
    def design_days(self, data):
        assert isinstance(data, list), 'Expected' \
            ' a list of design days. Got {}'.format(type(data))
        for item in data:
            assert isinstance(item, DesignDay), 'Expected' \
                ' DesignDay type. Got {}'.format(type(item))
        self._design_days = data
        for dd in self._design_days:
            if dd.location != self._location:
                dd.location = self._location
                print('Updating location of {} to {}.'.format(dd, self._location))

    def to_dict(self):
        """Convert the Design Day to a dictionary."""
        return {
            'type': 'DDY',
            'location': self.location.to_dict(),
            'design_days': [des_d.to_dict() for des_d in self.design_days]
        }

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()
    
    def __copy__(self):
        new_ddy = DDY(self._location, [dday.duplicate() for dday in self._design_days])
        new_ddy._file_path = self._file_path
        return new_ddy

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self._location),) + tuple(hash(dday) for dday in self._design_days)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, DDY) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __len__(self):
        return len(self._design_days)

    def __getitem__(self, key):
        return self._design_days[key]

    def __setitem__(self, key, value):
        assert isinstance(value, DesignDay), 'Expected' \
                ' DesignDay type. Got {}'.format(type(value))
        self._design_days[key] = value

    def __iter__(self):
        return iter(self._design_days)

    def __contains__(self, item):
        return item in self._design_days

    def __repr__(self):
        """DDY object representation."""
        return "DDY File - {} [# days: {}]".format(
            self.location.city, str(len(self._design_days)))



# coding=utf-8
"""Ladybug Header"""
from __future__ import division

from copy import deepcopy

from .analysisperiod import AnalysisPeriod
from .datatype.base import DataTypeBase


class Header(object):
    """DataCollection header.

    Header carries meatdata for DataCollections including data type, unit
    and analysis period.

    Args:
        data_type: A DataType object. (e.g. Temperature)
        unit: data_type unit (e.g. 'C').
        analysis_period: A Ladybug AnalysisPeriod object.
        metadata: Optional dictionary of additional metadata,
            containing information such as 'source', 'city', or 'zone'.

    Properties:
        * data_type
        * unit
        * analysis_period
        * metadata
    """

    __slots__ = ('_data_type', '_unit', '_analysis_period', '_metadata')

    def __init__(self, data_type, unit, analysis_period, metadata=None):
        """Initiate Ladybug header for lists.
        """
        assert isinstance(data_type, DataTypeBase), \
            'data_type must be a Ladybug DataType. Got {}'.format(type(data_type))
        data_type.is_unit_acceptable(unit)
        assert isinstance(analysis_period, AnalysisPeriod), \
            'analysis_period must be a Ladybug AnalysisPeriod. Got {}'.format(
                type(analysis_period))
        if metadata is not None:
            assert isinstance(metadata, dict), \
                'metadata must be a dictionary. Got {}'.format(type(metadata))

        self._data_type = data_type
        self._unit = unit
        self._analysis_period = analysis_period
        self._metadata = metadata or {}

    @classmethod
    def from_dict(cls, data):
        """Create a header from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

                {
                "data_type": {},  # Type of data (e.g. Temperature)
                "unit": "",  # string
                "analysis_period": {},  # A Ladybug AnalysisPeriod
                "metadata": {}  # A dictionary of metadata
                }
        """
        # assign default values
        assert 'data_type' in data, 'Required Header key "data_type" is missing!'
        assert 'unit' in data, 'Required Header key "unit" is missing!'
        assert 'analysis_period' in data, \
            'Required Header key "analysis_period" is missing!'

        data_type = DataTypeBase.from_dict(data['data_type'])
        ap = AnalysisPeriod.from_dict(data['analysis_period'])
        metadata = data['metadata'] if 'metadata' in data else None
        return cls(data_type, data['unit'], ap, metadata)

    @property
    def data_type(self):
        """A DataType object."""
        return self._data_type

    @property
    def unit(self):
        """A text string representing an abbreviated unit."""
        return self._unit

    @property
    def analysis_period(self):
        """A AnalysisPeriod object."""
        return self._analysis_period

    @property
    def metadata(self):
        """Dictionary of metadata associated with the Header."""
        return self._metadata

    def duplicate(self):
        """Return a copy of the header."""
        a_per = self.analysis_period.duplicate() if self.analysis_period else None
        return self.__class__(self.data_type, self.unit,
                              a_per, deepcopy(self.metadata))

    def to_tuple(self):
        """Return Ladybug header as a list."""
        return (
            self.data_type,
            self.unit,
            self.analysis_period,
            self.metadata
        )

    def __iter__(self):
        """Return data as tuple."""
        return iter(self.to_tuple())

    def to_dict(self):
        """Return a header as a dictionary."""
        a_per = self.analysis_period.to_dict() if self.analysis_period else None
        return {
            'data_type': self.data_type.to_dict(),
            'unit': self.unit,
            'analysis_period': a_per,
            'metadata': self.metadata,
            'type': 'Header',
        }

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __key(self):
        return (self.data_type, self.unit, self.analysis_period, self.metadata)

    def __eq__(self, other):
        return isinstance(other, Header) and self.__key() == other.__key()

    def __ne__(self, value):
        return not self.__eq__(value)

    def __repr__(self):
        """Return Ladybug header as a string."""
        a_per = self.analysis_period if self.analysis_period else ''
        if self.metadata != {}:
            meta_str = '\n'.join(['{}: {}'.format(key, val)
                                  for key, val in self.metadata.items()])
            return "{} ({})\n{}\n{}".format(
                self.data_type, self.unit, a_per, meta_str)
        else:
            return "{} ({})\n{}".format(
                self.data_type, self.unit, a_per)



# coding=utf-8
import ladybug.futil as futil

import pytest
import os


def test_download_file():
    """Test the download file capability"""
    wf_url = "https://www.energyplus.net/weather-download/" \
        "southwest_pacific_wmo_region_5/AUS//AUS_NSW.Sydney.947670_IWEC/all"
    wf_name = os.path.join("./tests/fixtures/zip", "AUS_NSW.Sydney.947670_IWEC.zip")
    futil.download_file(wf_url, wf_name)
    assert os.path.isfile(wf_name)
    assert 200000 < os.stat(wf_name).st_size < 300000
    os.remove(wf_name)


def test_unzip_file():
    """Test the unzip file capability"""
    folder = "./tests/fixtures/zip/extracted"
    wf_path = "./tests/fixtures/zip/test.zip"
    futil.unzip_file(wf_path, folder)
    extracted_epw = os.path.join(folder, "AUS_NSW.Sydney.947670_IWEC.epw")
    assert os.path.isfile(extracted_epw)
    assert 1500000 < os.stat(extracted_epw).st_size < 1600000
    futil.nukedir(folder)


def test_copy_files_to_folder():
    """Test the copy file to folder capability"""
    existing_file = "./tests/fixtures/ddy/chicago.ddy"
    target_dir = "./tests/fixtures/zip"
    futil.copy_files_to_folder([existing_file], target_dir)
    final_file = os.path.join(target_dir, "chicago.ddy")
    assert os.path.isfile(final_file)
    assert 20000 < os.stat(final_file).st_size < 30000
    os.remove(final_file)


def test_csv_to_matrix():
    """Test the csv_to_matrix functions."""
    path = './tests/fixtures/epw/tokyo.epw'
    epw_mtx = futil.csv_to_matrix(path)
    assert len(epw_mtx) == 8768

    with pytest.raises(Exception):
        epw_mtx = futil.csv_to_num_matrix(path)



# coding=utf-8
from __future__ import division

from ladybug.graphic import GraphicContainer
from ladybug.legend import Legend, LegendParameters
from ladybug.datatype.thermalcondition import PredictedMeanVote
from ladybug.datatype.temperature import Temperature

from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.mesh import Mesh3D
from ladybug_geometry.geometry2d.mesh import Mesh2D

import pytest


def test_init_graphic_con():
    """Test the initialization of GraphicContainer objects."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [0, 1, 2, 3]
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max)

    str(graphic_con)  # Test the GraphicContainer representation

    assert len(graphic_con) == 4
    assert graphic_con[0] == 0
    assert graphic_con[-1] == 3
    for item in graphic_con:
        assert isinstance(item, (float, int))

    assert len(graphic_con.values) == 4
    assert isinstance(graphic_con.legend, Legend)
    assert graphic_con.value_colors == graphic_con.legend.value_colors

    assert graphic_con.legend_parameters.is_base_plane_default
    assert graphic_con.legend_parameters.is_segment_height_default
    assert graphic_con.legend_parameters.is_segment_width_default
    assert graphic_con.legend_parameters.is_text_height_default
    assert graphic_con.legend_parameters.base_plane != Plane()

    assert isinstance(graphic_con.lower_title_location, Plane)
    assert isinstance(graphic_con.upper_title_location, Plane)
    assert graphic_con.lower_title_location != Plane()
    assert graphic_con.upper_title_location != Plane()


def test_to_from_dict():
    """Test the to/from dict methods."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [0, 1, 2, 3]
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max)

    graphic_con_dict = graphic_con.to_dict()
    new_graphic_con = GraphicContainer.from_dict(graphic_con_dict)
    assert new_graphic_con.to_dict() == graphic_con_dict


def test_init_graphic_con_invalid():
    """Test the initialization of GraphicContainer objects with invalid inputs."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [0, 1, 2, 3, 4]

    with pytest.raises(Exception):
        GraphicContainer(data, mesh3d.min, mesh3d.max,
                         data_type=Temperature(), unit='NotAUnit')


def test_init_graphic_con_vertex_based():
    """Test the initialization of ResultMesh objects with vertex-based input."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max)

    assert len(graphic_con) == 9
    assert graphic_con[0] == 0
    assert graphic_con[-1] == 8

    assert len(graphic_con.values) == 9
    assert isinstance(graphic_con.legend_parameters, LegendParameters)
    assert isinstance(graphic_con.legend, Legend)
    assert graphic_con.value_colors == graphic_con.legend.value_colors


def test_init_graphic_con_legend_parameters():
    """Test the initialization of ResultMesh objects with a LegendParameters."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [-1, 0, 1, 2]
    legend_par = LegendParameters(base_plane=Plane(o=Point3D(2, 2, 0)))
    legend_par.vertical = False
    legend_par.segment_height = 0.25
    legend_par.segment_width = 0.5
    legend_par.text_height = 0.15
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max, legend_par)

    assert not graphic_con.legend_parameters.is_base_plane_default
    assert not graphic_con.legend_parameters.is_segment_height_default
    assert not graphic_con.legend_parameters.is_segment_width_default
    assert not graphic_con.legend_parameters.is_text_height_default
    assert not graphic_con.legend_parameters.vertical
    assert graphic_con.legend_parameters.base_plane.o == Point3D(2, 2, 0)
    assert graphic_con.legend_parameters.segment_height == 0.25
    assert graphic_con.legend_parameters.segment_width == 0.5
    assert graphic_con.legend_parameters.text_height == 0.15


def test_init_graphic_con_data_type():
    """Test the initialization of GraphicContainer objects with a DataType."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [-1, 0, 1, 2]
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max,
                                   data_type=Temperature())

    assert not graphic_con.legend_parameters.is_title_default
    assert graphic_con.legend_parameters.title == 'C'

    legend_par = LegendParameters()
    legend_par.vertical = False
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max,
                                   legend_par, data_type=Temperature())

    assert not graphic_con.legend_parameters.is_title_default
    assert graphic_con.legend_parameters.title == 'Temperature (C)'


def test_init_graphic_con_data_type_ordinal():
    """Test the ResultMesh objects with a DataType with unit_descr."""
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = [-1, 0, 1, 2]
    graphic_con = GraphicContainer(data, mesh3d.min, mesh3d.max,
                                   data_type=PredictedMeanVote(), unit='PMV')

    assert graphic_con.legend_parameters.min == -3
    assert graphic_con.legend_parameters.max == 3
    assert graphic_con.legend_parameters.segment_count == 7
    assert not graphic_con.legend_parameters.is_title_default
    assert graphic_con.legend_parameters.title == 'PMV'
    assert graphic_con.legend.segment_text == ['Cold', 'Cool', 'Slightly Cool',
                                               'Neutral',
                                               'Slightly Warm', 'Warm', 'Hot']



# coding=utf-8
from ladybug.location import Location


def test_init():
    """Test if the command correctly creates location based on individual values"""
    # This is totally a real place! It's in Wales, check it out!
    city = 'Llanfairpwllgwyngyllgogerychwyrndrobwllllantysiliogogogoch'
    country = 'United Kingdom'
    latitude = 53.2225252
    longitude = -4.2211707
    time_zone = 1
    elevation = 12
    station_id = 'SomeInventedStation'
    source = 'SomeNoneExistentSource'

    loc = Location(city=city, country=country, latitude=latitude,
                   longitude=longitude, time_zone=time_zone,
                   elevation=elevation, station_id=station_id, source=source)

    str(loc)  # test the string representation
    assert loc.city == city
    assert loc.country == country
    assert loc.latitude == latitude
    assert loc.longitude == longitude
    assert loc.time_zone == time_zone
    assert loc.elevation == elevation
    assert loc.station_id == station_id
    assert loc.source == source
    assert loc.meridian == -15


def test_default_values():
    """Test if the command correctly creates a location."""
    loc = Location()
    assert loc.city == '-'
    assert loc.country == '-'
    assert loc.latitude == 0
    assert loc.longitude == 0
    assert loc.time_zone == 0
    assert loc.station_id is None
    assert loc.source is None


def test_from_individual_values():
    """Test if the command correctly creates location based on individual values"""
    loc = Location()

    new_latitude = 31.4234953
    new_longitude = 72.9492158
    new_time_zone = 5
    new_elevation = 20

    loc.latitude = new_latitude
    loc.longitude = new_longitude
    loc.time_zone = new_time_zone
    loc.elevation = new_elevation

    assert loc.latitude == new_latitude
    assert loc.longitude == new_longitude
    assert loc.time_zone == new_time_zone
    assert loc.elevation == new_elevation


def test_from_location():
    """Test the from_location() class method"""
    city = 'Tehran'
    country = 'Iran'
    latitude = 36
    longitude = 34
    time_zone = 3.5
    elevation = 54

    loc = Location(city=city, country=country, latitude=latitude,
                   longitude=longitude, time_zone=time_zone,
                   elevation=elevation)

    loc_from_loc = Location.from_location(loc)

    assert loc_from_loc.city == city
    assert loc_from_loc.country == country
    assert loc_from_loc.latitude == latitude
    assert loc_from_loc.longitude == longitude
    assert loc_from_loc.time_zone == time_zone
    assert loc_from_loc.elevation == elevation


def test_dict_methods():
    """Test JSON serialization functions"""
    city = 'Tehran'
    country = 'Iran'
    latitude = 36
    longitude = 34
    time_zone = 3.5
    elevation = 54

    loc = Location(city=city, country=country, latitude=latitude,
                   longitude=longitude, time_zone=time_zone,
                   elevation=elevation)

    assert loc.to_dict() == {"city": city, "state": '-', "country": country,
                             "latitude": latitude, "longitude": longitude,
                             "time_zone": time_zone, "elevation": elevation,
                             "station_id": None, "source": None, "type": 'Location'}

    loc_from_dict = Location.from_dict(loc.to_dict())

    assert loc_from_dict.city == city
    assert loc_from_dict.latitude == latitude
    assert loc_from_dict.longitude == longitude
    assert loc_from_dict.time_zone == time_zone
    assert loc_from_dict.elevation == elevation
    assert loc_from_dict.country == country


def test_idf_methods():
    """Test the to/from idf methods."""
    loc = Location('Tehran', '', 'Iran', 36, 34, 3.5, 54)
    loc_idf = loc.to_idf()
    new_idf = Location.from_idf(loc_idf)
    assert loc_idf == new_idf.to_idf()


def test_duplicate():
    """Test the duplicate method."""
    city = 'Tehran'
    country = 'Iran'
    latitude = 36
    longitude = 34
    time_zone = 3.5
    elevation = 54

    loc = Location(city=city, country=country, latitude=latitude,
                   longitude=longitude, time_zone=time_zone,
                   elevation=elevation)
    loc_dup = loc.duplicate()

    assert loc.city == loc_dup.city == city
    assert loc.country == loc_dup.country == country
    assert loc.latitude == loc_dup.latitude == latitude
    assert loc.longitude == loc_dup.longitude == longitude
    assert loc.time_zone == loc_dup.time_zone == time_zone
    assert loc.elevation == loc_dup.elevation == elevation



# coding=utf-8
"""Thermal condition data type."""
from __future__ import division

from .base import DataTypeBase


class ThermalCondition(DataTypeBase):
    """Thermal Condition
    """
    _units = ('condition', 'PMV')
    _si_units = ('condition', 'PMV')
    _ip_units = ('condition', 'PMV')
    _min = -1
    _max = 1
    _abbreviation = 'Tcond'
    _unit_descr = {-1: 'Cold', 0: 'Neutral', 1: 'Hot'}

    def _condition_to_PMV(self, value):
        return value

    def _PMV_to_condition(self, value):
        return value

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('condition', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        return values, from_unit

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        return values, from_unit


class PredictedMeanVote(ThermalCondition):
    _min = float('-inf')
    _max = float('+inf')
    _abbreviation = 'PMV'
    _unit_descr = {-3: 'Cold', -2: 'Cool', -1: 'Slightly Cool', 0: 'Neutral',
                   1: 'Slightly Warm', 2: 'Warm', 3: 'Hot'}


class ThermalComfort(ThermalCondition):
    _min = 0
    _max = 1
    _abbreviation = 'TC'
    _unit_descr = {1: 'Comfortable', 0: 'Uncomfortable'}


class DiscomfortReason(ThermalCondition):
    _min = -2
    _max = 2
    _abbreviation = 'RDiscomf'
    _unit_descr = {-2: 'Too Dry', -1: 'Too Cold', 0: 'Comfortable',
                   1: 'Too Hot', 2: 'Too Humid'}


class ThermalConditionFivePoint(ThermalCondition):
    _min = -2
    _max = 2
    _abbreviation = 'Tcond-5'
    _unit_descr = {-2: 'Strong/Extreme Cold', -1: 'Moderate Cold',
                   0: 'No Thermal Stress', 1: 'Moderate Heat', 2: 'Strong/Extreme Heat'}


class ThermalConditionSevenPoint(ThermalCondition):
    _min = -3
    _max = 3
    _abbreviation = 'Tcond-7'
    _unit_descr = {-3: 'Very Strong/Extreme Cold', -2: 'Strong Cold',
                   -1: 'Moderate Cold', 0: 'No Thermal Stress', 1: 'Moderate Heat',
                   2: 'Strong Heat', 3: 'Very Strong/Extreme Heat'}


class ThermalConditionNinePoint(ThermalCondition):
    _min = -4
    _max = 4
    _abbreviation = 'Tcond-9'
    _unit_descr = {-4: 'Very Strong/Extreme Cold', -3: 'Strong Cold',
                   -2: 'Moderate Cold', -1: 'Slight Cold', 0: 'No Thermal Stress',
                   1: 'Slight Heat', 2: 'Moderate Heat', 3: 'Strong Heat',
                   4: 'Very Strong/Extreme Heat'}


class ThermalConditionElevenPoint(ThermalCondition):
    _min = -5
    _max = 5
    _abbreviation = 'Tcond-11'
    _unit_descr = {-5: 'Extreme Cold', -4: 'Very Strong Cold', -3: 'Strong Cold',
                   -2: 'Moderate Cold', -1: 'Slight Cold', 0: 'No Thermal Stress',
                   1: 'Slight Heat', 2: 'Moderate Heat', 3: 'Strong Heat',
                   4: 'Very Strong Heat', 5: 'Extreme Heat'}


class UTCICategory(ThermalCondition):
    _min = 0
    _max = 9
    _abbreviation = 'UTCIcond'
    _unit_descr = {0: 'Extreme Cold Stress', 1: 'Very Strong Cold Stress',
                   2: 'Strong Cold Stress', 3: 'Moderate Cold Stress',
                   4: 'Slight Cold Stress', 5: 'No Thermal Stress',
                   6: 'Moderate Heat Stress', 7: 'Strong Heat Stress',
                   8: 'Strong Heat Stress', 9: 'Extreme Heat Stress'}



# coding=utf-8
"""Utilities for finding roots of continuous functions."""
from __future__ import division


def secant(a, b, fn, epsilon):
    """One of the fastest root-finding algorithms.
    The method calculates the slope of the function fn and this enables it to converge
    to a solution very fast. However, if started too far away from a root, the method
    may not converge (returning a None). For this reason, it is recommended that this
    function be used first in any guess-and-check workflow and, if it fails to find a
    root, the bisect() method should be used.

    Args:
        a: The lowest possible boundary of the value you are tying to find.
        b: The highest possible boundary of the value you are tying to find.
        fn: A function representing the relationship between the value you are
            trying to find and the target condition you are trying to satisfy.
            It should typically be structured in the following way:

            .. code-block:: python

                def fn(value_trying_to_find):
                    funct(value_trying_to_find) - target_desired_from_funct

            ...but the subtraction should be switched if value_trying_to_find
            has a negative relationship with the funct.
        epsilon: The acceptable error in the target_desired_from_funct.

    Returns:
        root -- The value that gives the target_desired_from_funct.

    References
    ----------
    [1] Wikipedia contributors. (2018, December 29). Root-finding algorithm.
    In Wikipedia, The Free Encyclopedia. Retrieved 18:16, December 30, 2018,
    from https://en.wikipedia.org/wiki/Root-finding_algorithm#Secant_method
    """
    f1 = fn(a)
    if abs(f1) <= epsilon:
        return a
    f2 = fn(b)
    if abs(f2) <= epsilon:
        return b
    for i in range(100):
        slope = (f2 - f1) / (b - a)
        c = b - f2 / slope
        f3 = fn(c)
        if abs(f3) < epsilon:
            return c
        a = b
        b = c
        f1 = f2
        f2 = f3

    return None


def bisect(a, b, fn, epsilon, target):
    """The simplest root-finding algorithm.

    It is extremely reliable. However, it converges slowly for this reason,
    it is recommended that this only be used after the secant() method has
    returned None.

    Args:
        a: A lower guess of the value you are tying to find.
        b: A higher guess of the value you are tying to find.
        fn: A function representing the relationship between the value you are
            trying to find and the target condition you are trying to satisfy.
            It should typically be structured in the following way:

            .. code-block:: python

                def fn(value_trying_to_find):
                    funct(value_trying_to_find) - target_desired_from_funct

            ...but the subtraction should be switched if value_trying_to_find
            has a negative relationship with the funct.
        epsilon: The acceptable error in the target_desired_from_funct.
        target: The target slope (typically 0 for a local minima or maxima).

    Returns:
        root -- The value that gives the target_desired_from_funct.

    References
    ----------
    [1] Wikipedia contributors. (2018, December 29). Root-finding algorithm.
    In Wikipedia, The Free Encyclopedia. Retrieved 18:16, December 30, 2018,
    from https://en.wikipedia.org/wiki/Root-finding_algorithm#Bisection_method
    """
    while (abs(b - a) > 2 * epsilon):
        midpoint = (b + a) / 2
        a_t = fn(a)
        b_t = fn(b)
        midpoint_t = fn(midpoint)
        if (a_t - target) * (midpoint_t - target) < 0:
            b = midpoint
        elif (b_t - target) * (midpoint_t - target) < 0:
            a = midpoint
        else:
            return -999

    return midpoint



# coding=utf-8
"""Energy intensity data type."""
from __future__ import division

from .base import DataTypeBase


class EnergyIntensity(DataTypeBase):
    """Energy Intensity
    """
    _units = ('kWh/m2', 'kBtu/ft2', 'Wh/m2', 'Btu/ft2', 'kWh/ft2', 'kBtu/m2')
    _si_units = ('kWh/m2', 'Wh/m2')
    _ip_units = ('kBtu/ft2', 'Btu/ft2')
    _abbreviation = 'EUI'
    _point_in_time = False
    _cumulative = True

    def _kWh_m2_to_kBtu_ft2(self, value):
        return value * 0.316998

    def _kWh_m2_to_Wh_m2(self, value):
        return value * 1000.

    def _kWh_m2_to_Btu_ft2(self, value):
        return value * 316.998

    def _kWh_m2_to_kWh_ft2(self, value):
        return value / 10.7639

    def _kWh_m2_to_kBtu_m2(self, value):
        return value * 3.41214

    def _kBtu_ft2_to_kWh_m2(self, value):
        return value / 0.316998

    def _Wh_m2_to_kWh_m2(self, value):
        return value / 1000.

    def _Btu_ft2_to_kWh_m2(self, value):
        return value / 316.998

    def _kWh_ft2_to_kWh_m2(self, value):
        return value * 10.7639

    def _kBtu_m2_to_kWh_m2(self, value):
        return value / 3.41214

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('kWh/m2', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'Wh/m2':
            return self.to_unit(values, 'Btu/ft2', from_unit), 'Btu/ft2'
        else:
            return self.to_unit(values, 'kBtu/ft2', from_unit), 'kBtu/ft2'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'Btu/ft2':
            return self.to_unit(values, 'Wh/m2', from_unit), 'Wh/m2'
        else:
            return self.to_unit(values, 'kWh/m2', from_unit), 'kWh/m2'


class Radiation(EnergyIntensity):
    _min = 0
    _abbreviation = 'Esolar'


class GlobalHorizontalRadiation(Radiation):
    _abbreviation = 'GHR'


class DirectNormalRadiation(Radiation):
    _abbreviation = 'DNR'


class DiffuseHorizontalRadiation(Radiation):
    _abbreviation = 'DHR'


class DirectHorizontalRadiation(Radiation):
    _abbreviation = 'DR'


class ExtraterrestrialHorizontalRadiation(Radiation):
    _abbreviation = 'HRex'


class ExtraterrestrialDirectNormalRadiation(Radiation):
    _abbreviation = 'DNRex'
